#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 13:12:50 2018

@author: prmiles
"""
# import required packages
import numpy as np

class ErrorVarianceEstimator:
    '''
    Estimate observation errors.

    Attributes:
        * :meth:`~update_error_variance`
        * :meth:`~gammar`
        * :meth:`~gammar_mt`
    '''
    def __init__(self):
        self.description = 'Estimate error variance.'

    def update_error_variance(self, sos, model):
        '''
        Update observation error variance.

        **Strategy:** Treat error variance :math:`\\sigma^2` as parameter to be sampled.

        **Definition:** The property that the prior and posterior distributions have the
        same parametric form is termed conjugacy.

        Starting from the likelihood function, it can be shown

        .. math::

            \\sigma^2|(\\nu, q) \sim \\text{Inv-Gamma}\\Big(\\frac{N_s + N}{2}, \\frac{N_s\\sigma_{s}^2+ SS_q}{2}\\Big)

        where :math:`N_s` and :math:`\\sigma_{s}^2` are shape and scaling parameters,
        :math:`N` is the number of observations, and :math:`SS_q` is the sum-of-squares error.

        .. note::

            The variables :math:`N_s` and :math:`\\sigma_{s}^2` correspond
            to :code:`N0` and :code:`S20` in the :class:`~.ModelSettings` class, respectively.

        Args:
            * **sos** (:class:`~numpy.ndarray`): Return argument from evaluation of sum-of-squares function.
            * **model** (:class:`~.ModelSettings`): MCMC model settings.
        '''
        N0 = model.N0
        S20 = model.S20
        N = model.N
        sigma2 = model.sigma2 # initializes it as array
        nsos = len(sos)
        
        for jj in range(0,nsos):
            sigma2[jj] = (self.gammar(1, 1, 0.5*(N0[jj]+N[jj]),
                          2*((N0[jj]*S20[jj]+sos[jj])**(-1))))**(-1)

        return sigma2
            
    def gammar(self, m, n, a, b = 1):
        '''
        Random deviates from gamma distribution.

        Returns a m x n matrix of random deviates from the Gamma
          distribution with shape parameter A and scale parameter B:

        .. math::

            p(x|A,B) = \\frac{B^{-A}}{\\Gamma(A)}*x^{A-1}*\\exp(-x/B)

        Args:
            * **m** (:py:class:`int`): Number of rows in return
            * **n** (:py:class:`int`): Number of columns in return
            * **a** (:py:class:`float`): Shape parameter
            * **b** (:py:class:`float`): Scaling parameter
        '''
        if a <= 0: # special case
            y = np.zeros([m,n])
            return y
        
        y = self.gammar_mt(m, n, a, b)
        return y
    
    def gammar_mt(self, m, n, a, b = 1):
        '''
        Wrapper routine for calculating random deviates from gamma distribution
        using method of Marsaglia and Tsang (2000) :cite:`marsaglia2000simple`.

        Args:
            * **m** (:py:class:`int`): Number of rows in return
            * **n** (:py:class:`int`): Number of columns in return
            * **a** (:py:class:`float`): Shape parameter
            * **b** (:py:class:`float`): Scaling parameter
        '''
        y = np.zeros([m,n])
        for jj in range(0,n):
            for ii in range(0,m):
                y[ii,jj] = self._gammar_mt1(a = a, b = b)
                
        return y
        
    def _gammar_mt1(self, a, b = 1):
        '''
        Calculates random deviate from gamma distribution using method of
        Marsaglia and Tsang (2000).

        Args:
            * **a** (:py:class:`float`): Shape parameter
            * **b** (:py:class:`float`): Scaling parameter
        '''
        if a < 1:
            y = self._gammar_mt1(1+a,b)*np.random.rand(1)**(a**(-1))
            return y
        else:
            d = a - 3**(-1)
            c = (9*d)**(-0.5)
            while 1:
                while 1:
                    x = np.random.randn(1)
                    v = 1.0 + c*x
                    if v > 0:
                        break
                    
                v = v**(3)
                u = np.random.rand(1)
                if u < 1.0 - 0.0331*x**(4):
                    break
                if np.log(u) < 0.5*x**2 + d*(1.0 - v + np.log(v)):
                    break
            
            y = b*d*v
            return y


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 26 08:48:00 2018

@author: prmiles
"""

from pymcmcstat.structures.ParameterSet import ParameterSet
from pymcmcstat.samplers.Metropolis import Metropolis
import test.general_functions as gf

import unittest
from mock import patch
import numpy as np

# --------------------------
# Evaluation
# --------------------------
class UnpackSet(unittest.TestCase):

    def test_unpack_set(self):
        CL = {'theta':1.0, 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        parset = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])

        MA = Metropolis()
        oldpar, ss, oldprior, sigma2 = MA.unpack_set(parset)
        NL = {'theta':oldpar, 'ss': ss, 'prior':oldprior, 'sigma2': sigma2}
        self.assertDictEqual(CL,NL)

# --------------------------
class EvaluateLikelihood(unittest.TestCase):
    @classmethod
    def setup_size_2(cls):
        MA = Metropolis()
        ss1 = np.array([1., 2.])
        ss2 = np.array([1.1, 2.4])
        newprior = np.array([0.,0.])
        oldprior = np.array([0.,0.])
        sigma2 = np.array([1.,1.])
        return MA, ss1, ss2, newprior, oldprior, sigma2
    
    def test_size_of_alpha_for_1d_nsos(self):
        MA = Metropolis()
        ss1 = np.array([1.])
        ss2 = np.array([1.1])
        newprior = np.array([0.])
        oldprior = np.array([0.])
        sigma2 = np.array([1.])
        alpha = MA.evaluate_likelihood_function(ss1, ss2, sigma2, newprior, oldprior)
        self.assertEqual(alpha.size, 1)
        
    def test_size_of_alpha_for_2d_nsos(self):
        MA, ss1, ss2, newprior, oldprior, sigma2 = self.setup_size_2()
        alpha = MA.evaluate_likelihood_function(ss1, ss2, sigma2, newprior, oldprior)
        self.assertEqual(alpha.size, 1)
        
    def test_likelihood_goes_up(self):
        MA, ss1, ss2, newprior, oldprior, sigma2 = self.setup_size_2()
        alpha1 = MA.evaluate_likelihood_function(ss1, ss2, sigma2, newprior, oldprior)
        
        ss3 = np.array([0.4, 0.5])
        alpha2 = MA.evaluate_likelihood_function(ss3, ss2, sigma2, newprior, oldprior)
        self.assertTrue(alpha1 < alpha2)
        
    def test_likelihood_goes_down(self):
        MA, ss1, ss2, newprior, oldprior, sigma2 = self.setup_size_2()
        alpha1 = MA.evaluate_likelihood_function(ss1, ss2, sigma2, newprior, oldprior)
        
        ss3 = np.array([1.4, 2.5])
        alpha2 = MA.evaluate_likelihood_function(ss3, ss2, sigma2, newprior, oldprior)
        self.assertTrue(alpha2 < alpha1)
        
    def test_alpha_value(self):
        MA, ss1, ss2, newprior, oldprior, sigma2 = self.setup_size_2()
        alpha1 = MA.evaluate_likelihood_function(ss1, ss2, sigma2, newprior, oldprior)
        self.assertTrue(np.allclose(alpha1, np.array([2.568050833375483])), msg = str('alpha = {}'.format(alpha1)))
               
# --------------------------
class RunMetropolisStep(unittest.TestCase):
    @classmethod
    def setup_rms(cls, CL):
        sos_object, prior_object, parameters = gf.setup_mcmc_case_mh()
        MS = Metropolis()
        R = np.array([[0.4, 0.2],[0, 0.3]])
        parset = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        accept, newset, outbound, npar_sample_from_normal = MS.run_metropolis_step(old_set = parset, parameters = parameters, R = R, prior_object = prior_object, sos_object = sos_object)
        
        return accept, outbound

    @patch('pymcmcstat.samplers.Metropolis.is_sample_outside_bounds', return_value = True)
    def test_run_step_outside_bounds(self, mock_1):
        CL = {'theta':1.0, 'ss': 1.0, 'prior': 0.0, 'sigma2': 1.0}
        accept, outbound = self.setup_rms(CL)
        self.assertEqual(outbound, 1, msg = 'outbound set to 1')
        self.assertEqual(accept, 0, msg = 'Not accepted because outside bounds')
        
    @patch('pymcmcstat.samplers.Metropolis.is_sample_outside_bounds', return_value = False)
    @patch('pymcmcstat.samplers.Metropolis.Metropolis.evaluate_likelihood_function', return_value = 1.1)
    def test_run_step_inside_bounds(self, mock_1, mock_2):
        CL = {'theta':1.0, 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        accept, outbound = self.setup_rms(CL)
        self.assertEqual(outbound, 0, msg = 'outbound set to 0')
        self.assertEqual(accept, 1, msg = 'Accepted because likelihood > 1')
        
    @patch('pymcmcstat.samplers.Metropolis.is_sample_outside_bounds', return_value = False)
    @patch('pymcmcstat.samplers.Metropolis.Metropolis.evaluate_likelihood_function', return_value = 0.5)
    @patch('numpy.random.rand', return_value = 0.4)
    def test_run_step_inside_bounds_test_accept(self, mock_1, mock_2, mock_3):
        CL = {'theta':1.0, 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        accept, outbound = self.setup_rms(CL)
        self.assertEqual(outbound, 0, msg = 'outbound set to 0')
        self.assertEqual(accept, 1, msg = 'Accepted because 0.5 > 0.4')
        
    @patch('pymcmcstat.samplers.Metropolis.is_sample_outside_bounds', return_value = False)
    @patch('pymcmcstat.samplers.Metropolis.Metropolis.evaluate_likelihood_function', return_value = 0.3)
    @patch('numpy.random.rand', return_value = 0.4)
    def test_run_step_inside_bounds_test_accept_fail(self, mock_1, mock_2, mock_3):
        CL = {'theta':1.0, 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        accept, outbound = self.setup_rms(CL)
        self.assertEqual(outbound, 0, msg = 'outbound set to 0')
        self.assertEqual(accept, 0, msg = 'Accepted because 0.3 < 0.4')


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 10:30:29 2018

@author: prmiles
"""
# import required packages
import numpy as np
from ..structures.ParameterSet import ParameterSet
from .utilities import sample_candidate_from_gaussian_proposal
from .utilities import is_sample_outside_bounds, set_outside_bounds
from .utilities import acceptance_test

class Metropolis:
    '''
    .. |br| raw:: html

        <br>

    Pseudo-Algorithm:

        #. Sample :math:`z_k \sim N(0,1)`
        #. Construct candidate :math:`q^* = q^{k-1} + Rz_k`
        #. Compute |br| :math:`\quad SS_{q^*} = \\sum_{i=1}^N[v_i-f_i(q^*)]^2`
        #. Compute |br| :math:`\quad \\alpha = \\min\\Big(1, e^{[SS_{q^*} - SS_{q^{k-1}}]/(2\sigma^2_{k-1})}\Big)`
        #. If :math:`u_{\\alpha} <~\\alpha,` |br|
            Set :math:`q^k = q^*,~SS_{q^k} = SS_{q^*}`
           Else
            Set :math:`q^k = q^{k-1},~SS_{q^k} = SS_{q^{k-1}}`

    Attributes:
        * :meth:`~acceptance_test`
        * :meth:`~run_metropolis_step`
        * :meth:`~unpack_set`
    '''
    # --------------------------------------------------------
    def run_metropolis_step(self, old_set, parameters, R, prior_object, sos_object):
        '''
        Run Metropolis step.

        Args:
            * **old_set** (:class:`~.ParameterSet`): Features of :math:`q^{k-1}`
            * **parameters** (:class:`~.ModelParameters`): Model parameters
            * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of parameter covariance matrix
            * **priorobj** (:class:`~.PriorFunction`): Prior function
            * **sosobj** (:class:`~.SumOfSquares`): Sum-of-Squares function

        Returns:
            * **accept** (:py:class:`int`): 0 - reject, 1 - accept
            * **newset** (:class:`~.ParameterSet`): Features of :math:`q^*`
            * **outbound** (:py:class:`int`): 1 - rejected due to sampling outside of parameter bounds
            * **npar_sample_from_normal** (:class:`~numpy.ndarray`): Latet random sample points
        '''
           
        # unpack oldset
        oldpar, ss, oldprior, sigma2 = self.unpack_set(old_set)
        
        # Sample new candidate from Gaussian proposal
        newpar, npar_sample_from_normal = sample_candidate_from_gaussian_proposal(npar = parameters.npar, oldpar = oldpar, R = R)
           
        # Reject points outside boundaries
        outsidebounds = is_sample_outside_bounds(newpar, parameters._lower_limits[parameters._parind[:]], parameters._upper_limits[parameters._parind[:]])
        if outsidebounds is True:
            # proposed value outside parameter limits
            newset = ParameterSet(theta = newpar, sigma2 = sigma2)
            newset, outbound = set_outside_bounds(next_set = newset)
            accept = False
        else:
            outbound = 0
            # prior SS for the new theta
            newprior = prior_object.evaluate_prior(newpar)
            # calculate sum-of-squares
            ss2 = ss # old ss
            ss1 = sos_object.evaluate_sos_function(newpar)
            # evaluate likelihood
            alpha = self.evaluate_likelihood_function(ss1, ss2, sigma2, newprior, oldprior)
            # make acceptance decision
            accept = acceptance_test(alpha)
                    
            # store parameter sets in objects
            newset = ParameterSet(theta = newpar, ss = ss1, prior = newprior, sigma2 = sigma2, alpha = alpha)
        
        return accept, newset, outbound, npar_sample_from_normal
    # --------------------------------------------------------
    @classmethod
    def unpack_set(cls, parset):
        '''
        Unpack parameter set

        Args:
            * **parset** (:class:`~.ParameterSet`): Parameter set to unpack

        Returns:
            * **theta** (:class:`~numpy.ndarray`): Value of sampled model parameters
            * **ss** (:class:`~numpy.ndarray`): Sum-of-squares error using sampled value
            * **prior** (:class:`~numpy.ndarray`): Value of prior
            * **sigma2** (:class:`~numpy.ndarray`): Error variance
        '''
        theta = parset.theta
        ss = parset.ss
        prior = parset.prior
        sigma2 = parset.sigma2
        return theta, ss, prior, sigma2
    # --------------------------------------------------------
    @classmethod
    def evaluate_likelihood_function(cls, ss1, ss2, sigma2, newprior, oldprior):
        '''
        Evaluate likelihood function:

        .. math::

            \\alpha = \\exp\\Big[-0.5\\Big(\sum\\Big(\\frac{ SS_{q^*} - SS_{q^{k-1}} }{ \\sigma_{k-1}^2 }\\Big) + p_1 - p_2\\Big)\\Big]

        Args:
            * **ss1** (:class:`~numpy.ndarray`): SS error from proposed candidate, :math:`q^*`
            * **ss2** (:class:`~numpy.ndarray`): SS error from previous sample point, :math:`q^{k-1}`
            * **sigma2** (:class:`~numpy.ndarray`): Error variance estimate from previous sample point, :math:`\\sigma_{k-1}^2`
            * **newprior** (:class:`~numpy.ndarray`): Prior for proposal candidate
            * **oldprior** (:class:`~numpy.ndarray`): Prior for previous sample

        Returns:
            * **alpha** (:py:class:`float`): Result of likelihood function
        '''
        alpha = np.exp(-0.5*(sum((ss1 - ss2)*(sigma2**(-1))) + newprior - oldprior))
        return sum(alpha)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 21 12:21:24 2018

@author: prmiles
"""

from pymcmcstat.plotting import MCMCPlotting as MP
import matplotlib.pyplot as plt
import numpy as np
import math

import unittest

# --------------------------
class PlotDensityPanel(unittest.TestCase):
    def test_basic_plot_features(self):
        npar = 3
        chains = np.random.random_sample(size = (100,npar))
        f = MP.plot_density_panel(chains = chains)
        for ii in range(npar):
            name = str('$p_{{{}}}$'.format(ii))
            self.assertEqual(f.axes[ii].get_xlabel(), name, msg = str('Should be {}'.format(name)))
            self.assertEqual(f.axes[ii].get_ylabel(), str('$\pi$({}$|M^{}$)'.format(name, '{data}')), msg = 'Should be posterior')
        self.assertEqual(f.get_figwidth(), 5.0, msg = 'Width is 5in')
        self.assertEqual(f.get_figheight(), 4.0, msg = 'Height is 4in')
        plt.close()

    def test_basic_plot_features_with_hist_on(self):
        npar = 3
        chains = np.random.random_sample(size = (100,npar))
        f = MP.plot_density_panel(chains = chains, hist_on = True)
        for ii in range(npar):
            name = str('$p_{{{}}}$'.format(ii))
            self.assertEqual(f.axes[ii].get_xlabel(), name, msg = str('Should be {}'.format(name)))
            self.assertEqual(f.axes[ii].get_ylabel(), str('$\pi$({}$|M^{}$)'.format(name, '{data}')), msg = 'Should be posterior')
        self.assertEqual(f.get_figwidth(), 5.0, msg = 'Width is 5in')
        self.assertEqual(f.get_figheight(), 4.0, msg = 'Height is 4in')
        plt.close()
        
# --------------------------
class PlotChainPanel(unittest.TestCase):
    def test_basic_plot_features_nsimu_lt_maxpoints(self):
        chains = np.random.random_sample(size = (100,2))
        f = MP.plot_chain_panel(chains = chains)
        x1, y1 = f.axes[0].lines[0].get_xydata().T
        x2, y2 = f.axes[1].lines[0].get_xydata().T
        self.assertTrue(np.array_equal(y1, chains[:,0]), msg = 'Expect y1 to match column 1')
        self.assertTrue(np.array_equal(y2, chains[:,1]), msg = 'Expect y2 to match column 2')
        self.assertEqual(f.axes[0].get_xlabel(), '', msg = 'Should be blank')
        self.assertEqual(f.axes[1].get_xlabel(), 'Iteration', msg = 'Should be Iteration')
        plt.close()
        
    def test_basic_plot_features_nsimu_gt_maxpoints(self):
        nsimu = 1000
        chains = np.random.random_sample(size = (nsimu,2))
        f = MP.plot_chain_panel(chains = chains)
        x1, y1 = f.axes[0].lines[0].get_xydata().T
        x2, y2 = f.axes[1].lines[0].get_xydata().T
        skip = int(math.floor(nsimu/500))
        self.assertTrue(np.array_equal(y1, chains[range(0,nsimu,skip),0]), msg = 'Expect y1 to match column 1')
        self.assertTrue(np.array_equal(y2, chains[range(0,nsimu,skip),1]), msg = 'Expect y2 to match column 2')
        self.assertEqual(f.axes[0].get_xlabel(), '', msg = 'Should be blank')
        self.assertEqual(f.axes[1].get_xlabel(), 'Iteration', msg = 'Should be Iteration')
        plt.close()
        
# --------------------------
class PlotHistogramPanel(unittest.TestCase):
    def test_basic_plot_features_nsimu_lt_maxpoints(self):
        npar = 3
        chains = np.random.random_sample(size = (100,npar))
        f = MP.plot_histogram_panel(chains = chains)
        for ii in range(npar):
            self.assertEqual(f.axes[ii].get_xlabel(), str('$p_{{{}}}$'.format(ii)), msg = str('Should be $p_{{{}}}$'.format(ii)))
            self.assertEqual(f.axes[ii].get_ylabel(), '', msg = 'Should be blank')
        plt.close()
        
# --------------------------
class PlotPairwiseCorrelationPanel(unittest.TestCase):
    def test_basic_plot_features_nsimu_lt_maxpoints(self):
        chains = np.random.random_sample(size = (100,3))
        f = MP.plot_pairwise_correlation_panel(chains = chains)
        x1, y1 = f.axes[0].lines[0].get_xydata().T
        x2, y2 = f.axes[1].lines[0].get_xydata().T
        x3, y3 = f.axes[1].lines[0].get_xydata().T
        
        self.assertTrue(np.array_equal(x1, chains[:,0]), msg = 'Expect x1 to match column 0')
        self.assertTrue(np.array_equal(y1, chains[:,1]), msg = 'Expect y1 to match column 1')
        
        self.assertTrue(np.array_equal(x2, chains[:,0]), msg = 'Expect x2 to match column 0')
        self.assertTrue(np.array_equal(y2, chains[:,2]), msg = 'Expect y2 to match column 2')
        
        self.assertTrue(np.array_equal(x3, chains[:,0]), msg = 'Expect x3 to match column 1')
        self.assertTrue(np.array_equal(y3, chains[:,2]), msg = 'Expect y3 to match column 2')
        
        for ai in f.axes:
            self.assertEqual(ai.get_xlabel(), '', msg = 'Should be blank')
            
        self.assertEqual(f.axes[0].get_title(),'$p_{0}$', msg = 'Expect $p_{0}$')
        self.assertEqual(f.axes[2].get_title(),'$p_{1}$', msg = 'Expect $p_{1}$')
        
        self.assertEqual(f.axes[0].get_ylabel(),'$p_{1}$', msg = 'Expect $p_{1}$')
        self.assertEqual(f.axes[1].get_ylabel(),'$p_{2}$', msg = 'Expect $p_{2}$')
            
        plt.close()
        
# --------------------------
class PlotChainMetrics(unittest.TestCase):
    def test_basic_plot_features(self):
        chains = np.random.random_sample(size = (100,1))
        f = MP.plot_chain_metrics(chain = chains, name = 'a1')
        x1, y1 = f.axes[0].lines[0].get_xydata().T
        self.assertTrue(np.array_equal(y1, chains[:,0]), msg = 'Expect y1 to match column 1')
        self.assertEqual(f.axes[0].get_xlabel(), 'Iterations', msg = 'Should be Iterations')
        self.assertEqual(f.axes[0].get_ylabel(), 'a1-chain', msg = 'Should be a1-chain')
        self.assertEqual(f.axes[1].get_xlabel(), 'a1', msg = 'Strings should match')
        self.assertEqual(f.axes[1].get_ylabel(), 'Histogram of a1-chain', msg = 'Strings should match')
        self.assertEqual(f.get_figwidth(), 7.0, msg = 'Width is 7in')
        self.assertEqual(f.get_figheight(), 5.0, msg = 'Height is 5in')
        plt.close()
        
    def test_figsize_plot_features(self):
        chains = np.random.random_sample(size = (100,1))
        f = MP.plot_chain_metrics(chain = chains, name = 'a1', figsizeinches = (10,2))
        x1, y1 = f.axes[0].lines[0].get_xydata().T
        self.assertTrue(np.array_equal(y1, chains[:,0]), msg = 'Expect y1 to match column 1')
        self.assertEqual(f.axes[0].get_xlabel(), 'Iterations', msg = 'Should be Iterations')
        self.assertEqual(f.axes[0].get_ylabel(), 'a1-chain', msg = 'Should be a1-chain')
        self.assertEqual(f.axes[1].get_xlabel(), 'a1', msg = 'Strings should match')
        self.assertEqual(f.axes[1].get_ylabel(), 'Histogram of a1-chain', msg = 'Strings should match')
        self.assertEqual(f.get_figwidth(), 10.0, msg = 'Width is 10in')
        self.assertEqual(f.get_figheight(), 2.0, msg = 'Height is 2in')
        plt.close()
        
# --------------------------
class Plot(unittest.TestCase):
    def test_plot_attributes(self):
        P = MP.Plot()
        check_these = ['plot_density_panel', 'plot_chain_panel', 'plot_pairwise_correlation_panel', 'plot_histogram_panel', 'plot_chain_metrics']
        for ct in check_these:
            self.assertTrue(hasattr(P, ct), msg = str('P contains method {}'.format(ct)))


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 10:15:37 2018

@author: prmiles
"""

class ParameterSet:
    '''
    Basic MCMC parameter set.

    **Description:** Storage device for passing parameter sets back and forth between sampling methods.

    Args:
        * **theta** (:class:`~numpy.ndarray`): Sampled values.
        * **ss** (:class:`~numpy.ndarray`): Sum-of-squares error(s).
        * **prior** (:class:`~numpy.ndarray`): Result from prior function.
        * **sigma2** (:class:`~numpy.ndarray`): Observation errors.
        * **alpha** (:py:class:`float`): Result from evaluating likelihood function.
    '''
    def __init__(self, theta = None, ss= None, prior = None, sigma2 = None, alpha = None):
        self.theta = theta
        self.ss = ss
        self.prior = prior
        self.sigma2 = sigma2
        self.alpha = alpha


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 31 05:12:38 2018

@author: prmiles
"""

from pymcmcstat.procedures.SumOfSquares import SumOfSquares
from pymcmcstat.procedures.ErrorVarianceEstimator import ErrorVarianceEstimator
import test.general_functions as gf
import unittest
import numpy as np

# --------------------------
class InitializeEVE(unittest.TestCase):

    def test_init_EVE(self):
        EVE = ErrorVarianceEstimator()
        self.assertTrue(hasattr(EVE, 'description'))
        
# --------------------------
class UpdateEVE(unittest.TestCase):
    
    def test_eve_update_with_nsos_1(self):
        model, __, parameters, data = gf.setup_mcmc()
        SOS = SumOfSquares(model = model, data = data, parameters = parameters)
        theta = np.array([2., 5.])
        ss = SOS.evaluate_sos_function(theta = theta)
        EVE = ErrorVarianceEstimator()
        sigma2 = EVE.update_error_variance(sos = ss, model = model)
        self.assertTrue(isinstance(sigma2, np.ndarray), msg = 'Expect array return.')
        self.assertEqual(sigma2.size, 1, msg = 'Size of array is 1')
        self.assertTrue(isinstance(sigma2[0], float), msg = 'Numerical result returned')
        
    def test_eve_update_with_nsos_2(self):
        model, options, parameters, data = gf.setup_mcmc()
        nsos = 2
        model._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        ss = np.array([0.1, 0.2])
        EVE = ErrorVarianceEstimator()
        sigma2 = EVE.update_error_variance(sos = ss, model = model)
        self.assertTrue(isinstance(sigma2, np.ndarray), msg = 'Expect array return.')
        self.assertEqual(sigma2.size, 2, msg = 'Size of array is 2')
        self.assertTrue(isinstance(sigma2[0], float), msg = 'Numerical result returned')
        
# --------------------------
class Gammar(unittest.TestCase):
    
    def test_gammar_for_a_0(self):
        EVE = ErrorVarianceEstimator()
        m = 4
        n = 3
        ret = EVE.gammar(m = m, n = 3, a = 0)
        self.assertTrue(np.array_equal(ret, np.zeros([m,n])), msg = 'Expect equal array return.')
        
    def test_gammar_for_a_lt_0(self):
        EVE = ErrorVarianceEstimator()
        m = 4
        n = 3
        ret = EVE.gammar(m = m, n = 3, a = -10.)
        self.assertTrue(np.array_equal(ret, np.zeros([m,n])), msg = 'Expect equal array return.')
        
# --------------------------
class Gammar_MT1(unittest.TestCase):
    
    def test_gammar_for_a_lt_1(self):
        EVE = ErrorVarianceEstimator()
        y = EVE._gammar_mt1(a = 0.1, b = 1)
        self.assertTrue(isinstance(y, np.ndarray), msg = 'Expect equal array return.')
        
#    def test_gammar_for_a_lt_0(self):
#        EVE = ErrorVarianceEstimator()
#        m = 4
#        n = 3
#        ret = EVE.gammar_mt1(a = -10.)
#        self.assertTrue(np.array_equal(ret, np.zeros([m,n])), msg = 'Expect equal array return.')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 30 07:34:40 2018

@author: prmiles
"""
from pymcmcstat.utilities.NumpyEncoder import NumpyEncoder
import unittest
import numpy as np

NE = NumpyEncoder()

# --------------------------
class NumpyEncoderTesting(unittest.TestCase):

    def test_NE_initialization_attributes(self):
        self.assertTrue(hasattr(NE, 'default'))

    def test_NE_return_ndarray(self):
        self.assertTrue(isinstance(NE.default(obj = np.ones([10,1])), list), msg = 'Numpy array converted to list')

    def test_NE_return_json(self):
        with self.assertRaises(TypeError, msg = 'dict not JSON serializable'):
            NE.default(obj = {'test_object': [10,1]})


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 17 09:03:37 2018

@author: prmiles
"""
# import required packages
import numpy as np

# -------------------------
# Define data structure
# -------------------------
class DataStructure:
    '''
    Structure for storing data in MCMC object.  
    The following random data sets will be referenced in examples for the 
    different class methods:
    ::

        x1 = np.random.random_sample(size = (5, 1))
        y1 = np.random.random_sample(size = (5, 2))

        x2 = np.random.random_sample(size = (10, 1))
        y2 = np.random.random_sample(size = (10, 3))

    Attributes:
        * :meth:`~add_data_set`
        * :meth:`~get_number_of_batches`
        * :meth:`~get_number_of_data_sets`
        * :meth:`~get_number_of_observations`
    '''
    def __init__(self):
        self.xdata = [] # initialize list
        self.ydata = [] # initialize list
        self.n = [] # initialize list - number of data points
        self.shape = [] # shape of ydata - important if information stored as matrix
        self.weight = [] # initialize list - weight of data set
        self.user_defined_object = [] # user defined object

    def add_data_set(self, x, y, n = None, weight = 1, user_defined_object = 0):
        '''
        Add data set to MCMC object.  
        
        This method must be called first before using any of the other methods 
        within :class:`~DataStructure`.
        ::

            mcstat = MCMC()
            mcstat.data.add_data_set(x = x1, y = y1)
            mcstat.data.add_data_set(x = x2, y = y2)

        This yields the following variables in the data structure.

        * `xdata` (:py:class:`list`): List of numpy arrays
            - :code:`xdata[0] = x1, xdata[0].shape = (5,1)`
            - :code:`xdata[1] = x2, xdata[1].shape = (10,1)`
        * `ydata` (:py:class:`list`): List of numpy arrays
            - :code:`ydata[0] = y1, ydata[0].shape = (5,2)`
            - :code:`ydata[1] = y2, ydata[1].shape = (10,3)`
        * `n` (:py:class:`list`): List of integers. :code:`n = [5, 10]`
        * `shape` (:py:class:`list`): List of `y.shape`. :code:`shape = [(5,2),(10,3)]`
        * `weight` (:py:class:`list`): List of weights. :code:`weight = [1, 1]`
        * `user_defined_object` (:py:class:`list`): List of objects. :code:`user_defined_object = [0,0]`

        Args:
            * **x** (:class:`~numpy.ndarray`): Independent data.  Recommend input as column vectors.
            * **y** (:class:`~numpy.ndarray`): Dependent data.  Recommend input as column vectors.
            * **n** (:py:class:`list`): List of integers denoting number of data points.
            * **weight** (:py:class:`list`): Weight of each data set.
            * **user_defined_object** (`User Defined`): Any object can be stored in this variable.

        .. note::
            In general, it is recommended that user's format their data as a column
            vector.  So, if you have `nds` independent data points, `x` and `y` should be
            `[nds,1]` or `[nds,]` numpy arrays.  Note if a list is sent, the code will
            convert it to a numpy array.
        '''
        
        # check that x and y are numpy arrays
        x = self._convert_to_numpy_array(x)
        y = self._convert_to_numpy_array(y)
        
        # convert to 2d arrays (if applicable)
        x = self._convert_numpy_array_to_2d(x)
        y = self._convert_numpy_array_to_2d(y)
        
        # append new data set
        self.xdata.append(x)
        self.ydata.append(y)
        
        if n is None:
            if isinstance(y, list): # y is a list
                self.n.append(len(y))
            elif isinstance(y, np.ndarray) and y.size == 1:
                self.n.append(y.size)
            else: # should
                self.n.append(y.shape[0]) # assume y is a numpy array - nrows is n
        
        self.shape.append(y.shape)
        
        self.weight.append(weight)
        # add user defined objects option
        self.user_defined_object.append(user_defined_object)
     
    @classmethod
    def _convert_to_numpy_array(cls, xy):
        '''
        Convert variable to numpy array.

        Args:
            * **xy** (`Unknown`): Variable to be converted

        Returns:
            * **xy** (:class:`~numpy.ndarray`): Variable as numpy array
        '''
        if isinstance(xy, np.ndarray) is False:
            xy = np.array([xy])
            
        return xy
    
    @classmethod
    def _convert_numpy_array_to_2d(cls, xy):
        '''
        Convert numpy array to at least 2d numpy array.

        Args:
            * **xy** (:class:`~numpy.ndarray`): Variable to be checked/converted

        Returns:
            * **xy** (:class:`~numpy.ndarray`): Variable as at least 2d numpy array
        '''
        if xy.ndim != 2: # numpy array is (xy.size,) -> Convert to (xy.size,1)
            xy = xy.reshape(xy.size,1)
            
        return xy
    
    def get_number_of_batches(self):
        '''
        Get number of batches in data structure.  Essentially, each time you call
        the :meth:`~add_data_set` method you are adding another batch. It is
        also equivalent to say the number of batches is equal to the length of the
        list `ydata`.  For example,
        ::

            nb = mcstat.data.get_number_of_batches()

        should return :code:`nb = 2` because :code:`len(mcstat.data.ydata) = 2`.

        Returns:
            * **nbatch** (:py:class:`int`): Number of batches.
        '''
        self.nbatch = len(self.shape)
        return self.nbatch
    
    def get_number_of_data_sets(self):
        '''
        Get number of data sets in data structure.  A data set is strictly
        speaking defined as the total number of columns in each element of
        the `ydata` list. For example,
        ::

            nds = mcstat.data.get_number_of_data_sets()

        should return :code:`nds = 2 + 3 = 5` because the number of columns in `y1` is
        2 and the number of columns in `y2` is 3.

        Returns:
            * Number of columns in `ydata` (:py:class:`int`)
        '''
        dshapes = self.shape
        ndatabatches = len(dshapes)
        nrows = []
        ncols = []
        for ii in range(ndatabatches):
            nrows.append(dshapes[ii][0])
            if len(dshapes[ii]) != 1:
                ncols.append(dshapes[ii][1])
        
        self.ndatasets = sum(ncols)
        
        return self.ndatasets
    
    def get_number_of_observations(self):
        '''
        Get number of observations in data structure.  An observation is essentially
        the total number of rows from each element of the `ydata` list. For example,
        ::

            nobs = mcstat.data.get_number_of_observations()

        should return :code:`nobs = 5 + 10 = 15` because the number of rows in `y1` is
        5 and the number of rows in `y2` is 10.

        Returns:
            * Number of rows in `ydata` (:class:`~numpy.ndarray`)
        '''
        n = np.sum(self.n)
        return np.array([n])


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Nov  8 12:00:11 2017

@author: prmiles
"""

import numpy as np
import sys
from ..settings.DataStructure import DataStructure
from ..settings.ModelSettings import ModelSettings
from ..utilities.progressbar import progress_bar
from ..plotting.utilities import append_to_nrow_ncol_based_on_shape, convert_flag_to_boolean, set_local_parameters
from ..plotting.utilities import empirical_quantiles, check_defaults
import matplotlib.pyplot as plt

class PredictionIntervals:
    '''
    Prediction/Credible interval methods.

    Attributes:
        - :meth:`~setup_prediction_interval_calculation`
        - :meth:`~generate_prediction_intervals`
        - :meth:`~plot_prediction_intervals`
    '''
    # ******************************************************************************
    # --------------------------------------------
    def setup_prediction_interval_calculation(self, results, data, modelfunction, burnin = 0):
        '''
        Setup calculation for prediction interval generation

        Args:
            * results (:class:`~.ResultsStructure`): MCMC results structure
            * data (:class:`~.DataStructure`): MCMC data structure
            * modelfunction: Model function handle
        '''
        # Analyze data structure
        self.__ndatabatches, ncols = self._analyze_data_structure(data = data)

        # setup data structure for prediction
        self.datapred = self._setup_data_structure_for_prediction(data = data, ndatabatches = self.__ndatabatches)

        # assign model function
        self.modelfunction = modelfunction

        # assign required features from the results structure
        self._assign_features_from_results_structure(results = results, burnin = burnin)

        # evaluate model function to determine shape of response
        self.__nrow, self.__ncol = self._determine_shape_of_response(modelfunction = modelfunction, ndatabatches = self.__ndatabatches, datapred = self.datapred, theta = self.__theta)

        # analyze structure of s2chain with respect to model output
        if self.__s2chain is not None:
            self.__s2chain_index = self._analyze_s2chain(ndatabatches = self.__ndatabatches, s2chain = self.__s2chain, ncol = self.__ncol)
    # --------------------------------------------
    @classmethod
    def _analyze_data_structure(cls, data):
        '''
        Analyze data structure.

        Args:
            * **data** (:class:`~.DataStructure`): Data structure.

        Returns:
            * **ndatabatches** (:py:class:`int`): Number of batch data sets.
            * **ncols** (:py:class:`list`): List containing number of columns in each set.
        '''
        # Analyze data structure
        dshapes = data.shape
        ndatabatches = len(dshapes)
        ncols = []
        for ii in range(ndatabatches):
            if len(dshapes[ii]) != 1:
                ncols.append(dshapes[ii][1])
            else:
                ncols.append(1)
        return ndatabatches, ncols
    # --------------------------------------------
    @classmethod
    def _setup_data_structure_for_prediction(cls, data, ndatabatches):
        '''
        Setup data structure for generating quantile.

        Args:
            * **data** (:class:`~.DataStructure`): Data structure.
            * **ndatabatches** (:py:class:`int`): Number of batch data sets.

        Returns:
            * **datapred** (:py:class:`list`): Data structure for interval generation.
        '''
        datapred = []
        for ii in range(ndatabatches):
            # setup data structure for prediction
            # this is required to allow user to send objects other than xdata to model function
            datapred.append(DataStructure())
            datapred[ii].add_data_set(x = data.xdata[ii], y = data.ydata[ii],
                         user_defined_object = data.user_defined_object[ii])
            
        return datapred
    # --------------------------------------------
    def _assign_features_from_results_structure(self, results, burnin = 0):
        '''
        Define variables based on items extracted from results dictionary.

        Args:
            * **results** (:py:class:`dict`): Results dictionary from MCMC simulation.
        '''
        # assign required features from the results structure
        self.__chain = results['chain'][burnin:,:]
        self.__s2chain = results['s2chain']
        if self.__s2chain is not None:
            self.__s2chain = self.__s2chain[burnin:,:]
            
        self.__parind = results['parind']
        self.__local = results['local']
        # Check how 'model' object was saved in results structure
        if isinstance(results['model_settings'], ModelSettings):
            self.__nbatch = results['model'].nbatch
        else:
            self.__nbatch = results['model_settings']['nbatch']
            
        self.__theta = results['theta']
        
        if 'sstype' in results:
            self.__sstype = results['sstype']
        else:
            self.__sstype = 0
    # --------------------------------------------
    @classmethod
    def _determine_shape_of_response(cls, modelfunction, ndatabatches, datapred, theta):
        '''
        Determine shape of model function repsonse.

        Args:
            * **modelfun** (:py:class:`func`): Model function handle.
        '''
        # evaluate model function to determine shape of response
        nrow = []
        ncol = []
        for ii in range(ndatabatches):
            if isinstance(modelfunction, list):
                y = modelfunction[ii](datapred[ii], theta)
            else:
                y = modelfunction(datapred[ii], theta)
                
            sh = y.shape
            nrow, ncol = append_to_nrow_ncol_based_on_shape(sh, nrow, ncol)
        return nrow, ncol

    # --------------------------------------------
    @classmethod
    def _analyze_s2chain(cls, ndatabatches, s2chain, ncol):
        '''
        Analysis of s2chain.

        Depending on the nature of the data structure, there are a couple ways
        to interpret the shape of the s2chain.  If s2chain = [nsimu, ns2], then
        ns2 could correspond to len(data.ydata) or the number of columns in
        data.ydata[0].  We will assume in this code that the user either stores
        similar sized vectors in a matrix; or, creates distinct list elements for
        equal or different size vectors.  Creating distinct list elements of matrices
        will generate an error when trying to plot prediction intervals.

        Args:
            * **ndatabatches** (:py:class:`int`): Number of batch data sets.
        '''
        # shape of s2chain
        n = s2chain.shape[1]
        
        # compare shape of s2chain with the number of data batches and the
        # number of columns in each batch.
        total_columns = sum(ncol)
        
        if n == 1: # only one obs. error for all data sets
            s2chain_index = np.zeros([ndatabatches,2], dtype = int)
            for ii in range(ndatabatches):
                s2chain_index[ii,:] = np.array([0, 1])
            
        elif n != 1 and total_columns == n: # then different obs. error for each column
            s2chain_index = np.zeros([ndatabatches,2], dtype = int)
            for ii in range(ndatabatches):
                if ii == 0:
                    s2chain_index[ii,:] = np.array([0, ncol[ii]])
                else:
                    s2chain_index[ii,:] = np.array([s2chain_index[ii-1,1], s2chain_index[ii-1,1] + ncol[ii]])
        elif n != 1 and total_columns != n:
            if n == ndatabatches: # assume separate obs. error for each batch
                s2chain_index = np.zeros([ndatabatches,2], dtype = int)
                for ii in range(ndatabatches):
                    if ii == 0: # 1?
                        s2chain_index[ii,:] = np.array([0, 1])
                    else:
                        s2chain_index[ii,:] = np.array([s2chain_index[ii-1,1], s2chain_index[ii-1,1] + 1])
            else:
                print('s2chain.shape = {}'.format(s2chain.shape))
                print('ndatabatches = {}'.format(ndatabatches))
                print('# of columns per batch = {}'.format(ncol))
                sys.exit('Unclear data structure: error variances do not match size of model output')
                
        return s2chain_index
    # ******************************************************************************
    # --------------------------------------------
    def generate_prediction_intervals(self, sstype = None, nsample = 500, calc_pred_int = True, waitbar = False):
        '''
        Generate prediction/credible interval.

        Args:
            * **sstype** (:py:class:`int`): Sum-of-squares type
            * **nsample** (:py:class:`int`): Number of samples to use in generating intervals.
            * **calc_pred_int** (:py:class:`bool`): Flag to turn on prediction interval calculation.
            * **waitbar** (:py:class:`bool`): Flag to turn on progress bar.
        '''
        
        chain, s2chain, lims, sstype, nsample, iisample = self._setup_generation_requirements(sstype = sstype, nsample = nsample, calc_pred_int = calc_pred_int)
        
        # setup progress bar
        print('Generating credible/prediction intervals:\n')
        if waitbar is True:
            self.__wbarstatus = progress_bar(iters = int(nsample))
            
        # extract chain elements
        testchain = chain[iisample,:]
            
        # calculate intervals for data sets
        if s2chain is None:
            credible_intervals = self._calculate_ci_for_data_sets(
                    testchain = testchain, waitbar = waitbar, lims = lims)
            prediction_intervals = None
        else:
            credible_intervals, prediction_intervals = self._calculate_ci_and_pi_for_data_sets(
                testchain = testchain, s2chain = s2chain, iisample = iisample, waitbar = waitbar, sstype = sstype, lims = lims)
                        
        # generate output dictionary
        self.intervals = {'credible_intervals': credible_intervals,
               'prediction_intervals': prediction_intervals}
    
        print('\nInterval generation complete\n')
    # --------------------------------------------
    def _setup_generation_requirements(self, nsample, calc_pred_int, sstype):
        '''
        Check if number of samples is defined - default to number of simulations.

        Args:
            * **nsample** (:py:class:`int`): Number of samples to draw from posterior.
            * **calc_pred_int** (:py:class:`bool`): User defined flag as to whether or not they want PI plotted.
            * **sstype** (:py:class:`int`): Flag to specify sstype.

        Returns:
            * **chain** (:class:`~numpy.ndarray`): Chain of posterior density.
            * **s2chain** (:class:`~numpy.ndarray`): Chain of observation errors.
            * **lims** (:class:`~numpy.ndarray`): Quantile limits.
            * **sstype** (:py:class:`int`): Flag to specify sstype.
            * **nsample** (:py:class:`int`): Number of samples to draw from posterior.
            * **iisample** (:class:`~numpy.ndarray`): Array of indices in posterior set.
        '''
        # extract chain & s2chain from results
        chain = self.__chain
        
        calc_pred_int = convert_flag_to_boolean(calc_pred_int)
        if calc_pred_int is False:
            s2chain = None
        else:
            s2chain = self.__s2chain
        
        # define number of simulations by the size of the chain array
        nsimu = chain.shape[0]
        
        # define interval limits
        lims = self._setup_interval_limits(s2chain)
        
        # define ss type
        sstype = self._setup_sstype(sstype)
        
        # check value of nsample
        nsample = self._check_nsample(nsample = nsample, nsimu = nsimu)
            
        # define sample points
        iisample, nsample = self._define_sample_points(nsample = nsample, nsimu = nsimu)
        
        return chain, s2chain, lims, sstype, nsample, iisample
    
    # --------------------------------------------
    @classmethod
    def _setup_interval_limits(cls, s2chain):
        '''
        Setup interval limits.

        If the observation errors are `None`, then it means prediction intervals
        are irrelevant.  As a default the routine will then plot an expanded
        set of credible intervals.  The output of this function should be
        interpreted as lower/upper limits of quantiles.  So, a 99% quantile will
        have the limits [0.005, ..., 0.995].

        Args:
            * **sigma2** (:class:`~numpy.ndarray` or `None`): Observation error chain.

        Returns:
            * **lims** (:class:`~numpy.ndarray`): Lower/Upper limits of quantiles.
        '''
        if s2chain is None:
            lims = np.array([0.005, 0.025, 0.05, 0.25, 0.5, 0.75, 0.9, 0.975, 0.995])
        else:
            lims = np.array([0.025, 0.5, 0.975])
        return lims
    # --------------------------------------------
    def _setup_sstype(self, sstype):
        '''
        Setup sstype and check value.

        :Cases:
            - `sstype = 0`: Standard normal.
            - `sstype = 1`: Square root.
            - `sstype = 2`: Logarithmic.

        See example usage in :meth:`~._observation_sample`.

        Args:
            * **sstype** (:py:class:`int`): Flag to specify sstype.

        Returns:
            * **sstype** (:py:class:`int`): Flag to specify sstype
        '''
        if sstype is None:
            sstype = self.__sstype
        else:
            sstype = 0
        return sstype
    # --------------------------------------------
    @classmethod
    def _check_nsample(cls, nsample, nsimu):
        '''
        Check if number of samples is defined - default to number of simulations.

        Args:
            * **nsample** (:py:class:`int`): Number of samples to draw from posterior.
            * **nsimu** (:py:class:`int`): Number of MCMC simulations.

        Returns:
            * **nsample** (:py:class:`int`): Number of samples to draw from posterior.
        '''
        # check value of nsample
        if nsample is None:
            nsample = nsimu
        return nsample
    # --------------------------------------------
    @classmethod
    def _define_sample_points(cls, nsample, nsimu):
        '''
        Define indices to sample from posteriors.

        Args:
            * **nsample** (:py:class:`int`): Number of samples to draw from posterior.
            * **nsimu** (:py:class:`int`): Number of MCMC simulations.

        Returns:
            * **iisample** (:class:`~numpy.ndarray`): Array of indices in posterior set.
            * **nsample** (:py:class:`int`): Number of samples to draw from posterior.
        '''
        # define sample points
        if nsample >= nsimu:
            iisample = range(nsimu) # sample all points from chain
            nsample = nsimu
        else:
            # randomly sample from chain
            iisample = np.ceil(np.random.rand(nsample)*nsimu) - 1
            iisample = iisample.astype(int)
        return iisample, nsample
    
    # --------------------------------------------
    def _calculate_ci_for_data_sets(self, testchain, waitbar, lims):
        '''
        Calculate credible intervals.

        Args:
            * **testchain** (:class:`~numpy.ndarray`): Sample points from posterior density.
            * **iisample** (:class:`~numpy.ndarray`): Array of indices in posterior set.
            * **waitbar** (:py:class:`bool`): Flag to turn on progress bar.
            * **sstype** (:py:class:`int`): Flag to specify sstype.

        Returns:
            * **credible_intervals(:py:class:`list`): List of credible intervals.
        '''
        credible_intervals = []
        for ii in range(len(self.datapred)):
            datapredii, nrow, ncol, modelfun, test = self._setup_interval_ii(ii = ii, datapred = self.datapred, nrow = self.__nrow, ncol = self.__ncol, modelfunction = self.modelfunction, local = self.__local)
            
            # Run interval generation on set ii
            ysave = self._calc_credible_ii(testchain = testchain, nrow = nrow, ncol = ncol,
                                     waitbar = waitbar, test = test, modelfun = modelfun, datapredii = datapredii)
                
            # generate quantiles
            plim = self._generate_quantiles(ysave, lims, ncol)
                
            credible_intervals.append(plim)
            
        return credible_intervals
    
    # --------------------------------------------
    def _calculate_ci_and_pi_for_data_sets(self, testchain, s2chain, iisample, waitbar, sstype, lims):
        '''
        Calculate prediction/credible intervals.

        Args:
            * **testchain** (:class:`~numpy.ndarray`): Sample points from posterior density.
            * **s2chain** (:class:`~numpy.ndarray`): Chain of observation errors.
            * **iisample** (:class:`~numpy.ndarray`): Array of indices in posterior set.
            * **waitbar** (:py:class:`bool`): Flag to turn on progress bar.
            * **sstype** (:py:class:`int`): Flag to specify sstype.

        Returns:
            * **credible_intervals(:py:class:`list`): List of credible intervals.
            * **prediction_intervals(:py:class:`list`): List of prediction intervals.
        '''
        credible_intervals = []
        prediction_intervals = []
        for ii in range(len(self.datapred)):
            datapredii, nrow, ncol, modelfun, test = self._setup_interval_ii(
                    ii = ii, datapred = self.datapred, nrow = self.__nrow, ncol = self.__ncol,
                    modelfunction = self.modelfunction, local = self.__local)
            s2ci = [self.__s2chain_index[ii][0], self.__s2chain_index[ii][1]]
            tests2chain = s2chain[iisample, s2ci[0]:s2ci[1]]
            
            # Run interval generation on set ii
            ysave, osave = self._calc_credible_and_prediction_ii(
                    testchain = testchain, tests2chain = tests2chain, nrow = nrow, ncol = ncol,
                    waitbar = waitbar, sstype = sstype, test = test, modelfun = modelfun, datapredii = datapredii)
                
            # generate quantiles
            plim = self._generate_quantiles(ysave, lims, ncol)
            olim = self._generate_quantiles(osave, lims, ncol)
                
            credible_intervals.append(plim)
            prediction_intervals.append(olim)
            
        return credible_intervals, prediction_intervals
    # --------------------------------------------
    @classmethod
    def _setup_interval_ii(cls, ii, datapred, nrow, ncol, modelfunction, local):
        '''
        Setup value for interval ii.

        Args:
            * **ii** (:py:class:`int`): Iteration number.
            * **datapred** (:py:class:`list`): List of data sets.
            * **nrow** (:py:class:`list`): List of rows in each data set.
            * **ncol** (:py:class:`list`): List of columns in each data set.
            * **modelfun** (:py:class:`func` or :py:class:`list`): Model function handle.

        Returns:
            * **datapredii** (:class:`~numpy.ndarray`): Data set.
            * **nrow** (:py:class:`int`): Number of rows in data set.
            * **ncol** (:py:class:`int`): Number of columns in data set.
            * **modelfun** (:py:class:`func`): Model function handle.
            * **test** (:class:`~numpy.ndarray`): Array of booleans correponding to local test.
        '''
        datapredii = datapred[ii]
        nrow = nrow[ii]
        ncol = ncol[ii]
        if isinstance(modelfunction, list):
            modelfun = modelfunction[ii]
        else:
            modelfun = modelfunction
        
        # some parameters may only apply to certain batch sets
        test = set_local_parameters(ii = ii, local = local)
        return datapredii, nrow, ncol, modelfun, test
    
    # --------------------------------------------
    def _calc_credible_ii(self, testchain, nrow, ncol, waitbar, test, modelfun, datapredii):
        '''
        Calculate response for set ii.

        Args:
            * **testchain** (:class:`~numpy.ndarray`): Sample points from posterior density.
            * **nrow** (:py:class:`int`): Number of rows in data set.
            * **ncol** (:py:class:`int`): Number of columns in data set.
            * **waitbar** (:py:class:`bool`): Flag to turn on progress bar.
            * **test** (:class:`~numpy.ndarray`): Array of booleans correponding to local test.
            * **modelfun** (:py:class:`func`): Model function handle.
            * **datapredii** (:class:`~numpy.ndarray`): Data set.

        Returns:
            * **ysave** (:class:`~numpy.ndarray`): Model responses.
        '''
        nsample = testchain.shape[0]
        theta = self.__theta
        ysave = np.zeros([nsample, nrow, ncol])
        
        for kk, isa in enumerate(testchain):
            # progress bar
            if waitbar is True:
                self.__wbarstatus.update(kk)
            
            # extract chain set
            theta[self.__parind[:]] = isa
            th = theta[test]
            # evaluate model
            ypred = modelfun(datapredii, th)
            ypred = ypred.reshape(nrow, ncol)
               
            # store model prediction
            ysave[kk,:,:] = ypred # store model output
        return ysave
    # --------------------------------------------
    def _calc_credible_and_prediction_ii(self, testchain, tests2chain, nrow, ncol, waitbar, sstype, test, modelfun, datapredii):
        '''
        Calculate response and observations for set ii.

        Args:
            * **testchain** (:class:`~numpy.ndarray`): Sample points from posterior density.
            * **tests2chain** (:class:`~numpy.ndarray`): Sample points from observation errors.
            * **nrow** (:py:class:`int`): Number of rows in data set.
            * **ncol** (:py:class:`int`): Number of columns in data set.
            * **waitbar** (:py:class:`bool`): Flag to turn on progress bar.
            * **sstype** (:py:class:`int`): Flag to specify sstype.
            * **test** (:class:`~numpy.ndarray`): Array of booleans correponding to local test.
            * **modelfun** (:py:class:`func`): Model function handle.
            * **datapredii** (:class:`~numpy.ndarray`): Data set.

        Returns:
            * **ysave** (:class:`~numpy.ndarray`): Model responses.
            * **osave** (:class:`~numpy.ndarray`): Model responses with observation errors.
        '''
        nsample = testchain.shape[0]
        theta = self.__theta
        ysave = np.zeros([nsample, nrow, ncol])
        osave = np.zeros([nsample, nrow, ncol])
        
        for kk, isa in enumerate(testchain):
            # progress bar
            if waitbar is True:
                self.__wbarstatus.update(kk)
            
            # extract chain set
            theta[self.__parind[:]] = isa
            th = theta[test]
            # evaluate model
            ypred = modelfun(datapredii, th)
            ypred = ypred.reshape(nrow, ncol)
            
            s2elem = tests2chain[kk]
            if s2elem.shape != (1,s2elem.size):
                s2elem = s2elem.reshape(1,s2elem.shape[0]) # make row vector
            opred = self._observation_sample(s2elem, ypred, sstype)
               
            # store model prediction
            ysave[kk,:,:] = ypred # store model output
            osave[kk,:,:] = opred # store model output with observation errors
        return ysave, osave
    # --------------------------------------------
    @classmethod
    def _observation_sample(cls, s2elem, ypred, sstype):
        '''
        Calculate model response with observation errors.

        Args:
            * **s2elem** (:class:`~numpy.ndarray`): Observation error(s).
            * **ypred** (:class:`~numpy.ndarray`): Model responses.
            * **sstype** (:py:class:`int`): Flag to specify sstype.

        Returns:
            * **opred** (:class:`~numpy.ndarray`): Model responses with observation errors.
        '''
        # check shape of s2elem and ypred
        ny = ypred.shape[1]
        ns = s2elem.shape[1]
        if ns != ny and ns == 1:
            s2elem = s2elem*np.ones([ny,1])
        elif ns != ny and ns != 1:
            sys.exit('Unclear data structure: error variances do not match size of model output')
            
        if sstype == 0:
            opred = ypred + np.matmul(np.random.standard_normal(ypred.shape),np.diagflat(
                    np.sqrt(s2elem))).reshape(ypred.shape)
        elif sstype == 1: # sqrt
            opred = (np.sqrt(ypred) + np.matmul(np.random.standard_normal(ypred.shape),np.diagflat(
                np.sqrt(s2elem))).reshape(ypred.shape))**2
        elif sstype == 2: # log
            opred = ypred*np.exp(np.matmul(np.random.standard_normal(ypred.shape),np.diagflat(
                np.sqrt(s2elem))).reshape(ypred.shape))
        else:
            sys.exit('Unknown sstype')
            
        return opred
    # --------------------------------------------
    @classmethod
    def _generate_quantiles(cls, response, lims, ncol):
        '''
        Generate quantiles based on observations.

        Args:
            * **response** (:class:`~numpy.ndarray`): Array of model responses.
            * **lims** (:class:`~numpy.ndarray`): Array of quantile limits.
            * **ncol** (:py:class:`int`): Number of columns in `ysave`.

        Returns:
            * **quantiles** (:py:class:`list`): Quantiles for intervals.
        '''
        # generate quantiles
        quantiles = []
        for jj in range(ncol):
            quantiles.append(empirical_quantiles(response[:,:,jj], lims))
            
        return quantiles
    
    # ******************************************************************************
    # --------------------------------------------
    def plot_prediction_intervals(self, plot_pred_int = True, adddata = False, addlegend = True, figsizeinches = None, model_display = {}, data_display = {}, interval_display = {}):
        '''
        Plot prediction/credible intervals.

        Args:
            * **plot_pred_int** (:py:class:`bool`): Flag to include PI on plot.
            * **adddata** (:py:class:`bool`): Flag to include data on plot.
            * **addlegend** (:py:class:`bool`): Flag to include legend on plot.
            * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height].
            * **model_display** (:py:class:`dict`): Model display settings.
            * **data_display** (:py:class:`dict`): Data display settings.
            * **interval_display** (:py:class:`dict`): Interval display settings.
            
        Available display options (defaults in parantheses):
            * **model_display**: `linestyle` (:code:`'-'`), `marker` (:code:`''`), `color` (:code:`'r'`), `linewidth` (:code:`2`), `markersize` (:code:`5`), `label` (:code:`model`), `alpha` (:code:`1.0`)
            * **data_display**: `linestyle` (:code:`''`), `marker` (:code:`'.'`), `color` (:code:`'b'`), `linewidth` (:code:`1`), `markersize` (:code:`5`), `label` (:code:`data`), `alpha` (:code:`1.0`)
            * **data_display**: `linestyle` (:code:`':'`), `linewidth` (:code:`1`), `alpha` (:code:`1.0`), `edgecolor` (:code:`'k'`)
        '''
        # unpack dictionary
        credible_intervals = self.intervals['credible_intervals']
        prediction_intervals = self.intervals['prediction_intervals']
        
        prediction_intervals, figsizeinches, nbatch, nn, clabels, plabels = self._setup_interval_plotting(
                plot_pred_int, prediction_intervals, credible_intervals, figsizeinches)
        
        # setup display settings
        interval_display, model_display, data_display = self._setup_display_settings(interval_display, model_display, data_display)
        
        # Define colors
        cicolor, picolor = self._setup_interval_colors(nn = nn, prediction_intervals = prediction_intervals)
        
        # initialize figure handle
        fighandle = []
        axhandle = []
        for ii in range(self.__ndatabatches):
                          
            credlims = credible_intervals[ii] # should be ny lists inside
            ny = len(credlims)
            
            # extract data
            dataii = self.datapred[ii]
            
            # define independent data
            time = dataii.xdata[0].reshape(dataii.xdata[0].shape[0],)
            
            for jj in range(ny):
                htmp, ax = self._initialize_plot_features(ii = ii, jj = jj, ny = ny, figsizeinches = figsizeinches)
                fighandle.append(htmp)
                axhandle.append(ax)

                # add prediction intervals - if applicable
                if prediction_intervals is not None:
                    ax.fill_between(time, prediction_intervals[ii][jj][0], prediction_intervals[ii][jj][-1],
                                    facecolor = picolor[0], label = plabels[0], **interval_display)
                
                # add range of credible intervals - if applicable
                for kk in range(0,int(nn)-1):
                    ax.fill_between(time, credlims[jj][kk], credlims[jj][-kk - 1], facecolor = cicolor[kk], label = clabels[kk], **interval_display)
                    
                # add model (median parameter values)
                ax.plot(time, credlims[jj][int(nn)-1], **model_display)
                
                # add data to plot
                if adddata is True:
                    plt.plot(dataii.xdata[0], dataii.ydata[0][:,jj], **data_display)
                    
                # add title
                self._add_batch_column_title(nbatch, ny, ii, jj)
                    
                # add legend
                if addlegend is True:
                    handles, labels = ax.get_legend_handles_labels()
                    ax.legend(handles, labels, loc = 'upper left')
    
        return fighandle, axhandle
    
    @classmethod
    def _setup_display_settings(cls, interval_display, model_display, data_display):
        '''
        Compare user defined display settings with defaults and merge.
        
        Args:
            * **interval_display** (:py:class:`dict`): User defined settings for interval display.
            * **model_display** (:py:class:`dict`): User defined settings for model display.
            * **data_display** (:py:class:`dict`): User defined settings for data display.

        Returns:
            * **interval_display** (:py:class:`dict`): Settings for interval display.
            * **model_display** (:py:class:`dict`): Settings for model display.
            * **data_display** (:py:class:`dict`): Settings for data display.
        '''
        # Setup interval display
        default_interval_display = {'linestyle': ':', 'linewidth': 1, 'alpha': 0.5, 'edgecolor': 'k'}
        interval_display = check_defaults(interval_display, default_interval_display)
        # Setup model display
        default_model_display = {'linestyle': '-', 'marker': '', 'color': 'r', 'linewidth': 2, 'markersize': 5, 'label': 'model', 'alpha': 1.0}
        model_display = check_defaults(model_display, default_model_display)
        # Setup data display
        default_data_display = {'linestyle': '', 'marker': '.', 'color': 'b', 'linewidth': 1, 'markersize': 5, 'label': 'data', 'alpha': 1.0}
        data_display = check_defaults(data_display, default_data_display)
        return interval_display, model_display, data_display
        
    @classmethod
    def _setup_interval_colors(cls, nn, prediction_intervals = None):
        ci = []
        pi = []
        if prediction_intervals is not None:
            ci.append('#c7e9b4')
            pi.append('#225ea8')
        else:
            ci.append('#253494')
            ci.append('#1d91c0')
            ci.append('#7fcdbb')
            ci.append('#c7e9b4')
        return ci, pi
    
    @classmethod
    def _initialize_plot_features(cls, ii, jj, ny, figsizeinches):
        '''
        Initialize plot for prediction/credible intervals.

        Args:
            * **ii** (:py:class:`int`): Batch #
            * **jj** (:py:class:`int`): Column #
            * **ny** (:py:class:`ny`): Number of intervals
            * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height].
        '''
        fighandbatch = str('Batch # {}'.format(ii))
        fighandcolumn = str('Column # {}'.format(jj))
        fighand = str('{} | {}'.format(fighandbatch, fighandcolumn))
        htmp = plt.figure(fighand, figsize=(figsizeinches)) # create new figure
        
        plt.figure(fighand)
        ax = plt.subplot(ny,1,jj+1)
        plt.tight_layout()
        plt.subplots_adjust(hspace=0.5)
        
        return htmp, ax
     
    @classmethod
    def _add_batch_column_title(cls, nbatch, ny, ii, jj):
        '''
        Add title to plot based on batch/column number.

        Args:
            * **nbatch** (:py:class:`nbatch`): Number of batches
            * **ny** (:py:class:`ny`): Number of intervals
            * **ii** (:py:class:`int`): Batch #
            * **jj** (:py:class:`int`): Column #
        '''
        # add title
        if nbatch > 1:
            plt.title(str('Batch #{}, Column #{}'.format(ii,jj)))
        elif ny > 1:
            plt.title(str('Column #{}'.format(jj)))
                          
    # --------------------------------------------
    def _setup_interval_plotting(self, plot_pred_int, prediction_intervals, credible_intervals, figsizeinches):
        '''
        Setup variables for interval plotting

        Args:
            * **plot_pred_int** (:py:class:`bool`): Flag to plot prediction interval
            * **prediction_intervals** (:py:class:`list`): Prediction intervals
            * **credible_intervals** (:py:class:`list`): Credible intervals
            * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height].

        Returns:
            * **prediction_intervals** (:py:class:`list` or `None`): Prediction intervals
            * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height].
            * **nbatch** (:py:class:`int`): Number of batches
            * **nn** (:py:class:`int`): Line number corresponding to median.
            * **clabels** (:py:class:`list`): List of label strings for credible intervals.
            * **plabels** (:py:class:`list`): List of label strings for prediction intervals.
        '''
        prediction_intervals = self._check_prediction_interval_flag(plot_pred_int = plot_pred_int, prediction_intervals = prediction_intervals)
            
        # check if figure size was specified
        if figsizeinches is None:
            figsizeinches = [7,5]
            
        nbatch, nn, nlines = self._setup_counting_metrics(credible_intervals = credible_intervals)
        
        clabels, plabels = self._setup_labels(prediction_intervals = prediction_intervals, nlines = nlines)
        return prediction_intervals, figsizeinches, nbatch, nn, clabels, plabels
    
    # --------------------------------------------
    @classmethod
    def _check_prediction_interval_flag(cls, plot_pred_int, prediction_intervals):
        '''
        Check prediction interval flag.

        Args:
            * **plot_pred_int** (:py:class:`bool`): Flag to plot prediction interval
            * **prediction_intervals** (:py:class:`list`): Prediction intervals

        Returns:
            * **prediction_intervals** (:py:class:`list` or `None`): Prediction intervals
        '''
        # check if prediction intervals exist and if user wants to plot them
        plot_pred_int = convert_flag_to_boolean(plot_pred_int)
        if plot_pred_int is False or prediction_intervals is None:
            prediction_intervals = None # turn off prediction intervals
        return prediction_intervals
    # --------------------------------------------
    @classmethod
    def _setup_labels(cls, prediction_intervals, nlines):
        '''
        Setup labels for prediction/credible intervals.

        Args:
            * **prediction_intervals** (:py:class:`list`): Prediction intervals
            * **nlines** (:py:class:`int`): Number of lines

        Returns:
            * **clabels** (:py:class:`list`): List of label strings for credible intervals.
            * **plabels** (:py:class:`list`): List of label strings for prediction intervals.
        '''
        clabels = ['95% CI']
        plabels = ['95% PI']
        
        # check if prediction intervals exist
        if prediction_intervals is None:
            clabels = ['99% CI', '95% CI', '90% CI', '50% CI']
            
        if prediction_intervals is None and nlines == 1:
            clabels = ['95% CI']
            
        return clabels, plabels
    # --------------------------------------------
    @classmethod
    def _setup_counting_metrics(cls, credible_intervals):
        '''
        Setup counting metrics for prediction/credible intervals.

        Args:
            * **credible_intervals** (:py:class:`list`): Credible intervals

        Returns:
            * **nbatch** (:py:class:`int`): Number of batches
            * **nn** (:py:class:`int`): Line number corresponding to median.
            * **nlines** (:py:class:`int`): Number of lines
        '''
        # define number of batches
        nbatch = len(credible_intervals)
        
        # define counting metrics
        nlines = len(credible_intervals[0][0]) # number of lines
        nn = np.int((nlines + 1)/2) # median
        nlines = nn - 1
        
        return nbatch, nn, nlines


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  4 10:22:54 2018

@author: prmiles
"""

from pymcmcstat.chain import ChainStatistics
import unittest
import numpy as np

CS = ChainStatistics
chain = np.random.random_sample(size = (1000,2))
chain[:,1] = 1e6*chain[:,1]

# --------------------------
# chainstats
# --------------------------
class Chainstats_Eval(unittest.TestCase):
    
    def test_cs_eval_with_return(self):
        stats = CS.chainstats(chain = chain, returnstats = True)
        self.assertTrue(isinstance(stats,dict))
        
    def test_cs_eval_with_no_return(self):
        stats = CS.chainstats(chain = chain, returnstats = False)
        self.assertEqual(stats, None)
        
    def test_cs_eval_with_no_chain(self):
        stats = CS.chainstats(chain = None, returnstats = True)
        self.assertTrue(isinstance(stats, str))
        
class BatchMeanSTD(unittest.TestCase):
    
    def test_len_s(self):
        s = CS.batch_mean_standard_deviation(chain, b = None)
        self.assertEqual(len(s), 2)
        
    def test_too_few_batches(self):
        with self.assertRaises(SystemExit, msg = 'too few batches'):
            CS.batch_mean_standard_deviation(chain, b = chain.shape[0])
            
class PowerSpectralDensity(unittest.TestCase):
    
    def test_nfft_none_size(self):
        x = chain[:,0]
        y = CS.power_spectral_density_using_hanning_window(x = x)
        nfft = min(len(x),256)
        n2 = int(np.floor(nfft/2))
        self.assertEqual(n2, len(y))
        
    def test_nfft_not_none_size(self):
        x = chain[:,0]
        nfft = 100
        y = CS.power_spectral_density_using_hanning_window(x = x, nfft = nfft)
        n2 = int(np.floor(nfft/2))
        self.assertEqual(n2, len(y))
        
    def test_nw_not_none(self):
        x = chain[:,0]
        y = CS.power_spectral_density_using_hanning_window(x = x, nw = len(x))
        nfft = min(len(x),256)
        n2 = int(np.floor(nfft/2))
        self.assertEqual(n2, len(y))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  3 14:30:05 2018

@author: prmiles
"""

from pymcmcstat.ParallelMCMC import ParallelMCMC
from pymcmcstat.ParallelMCMC import check_options_output, check_directory, check_initial_values
from pymcmcstat.ParallelMCMC import run_serial_simulation, assign_number_of_cores, generate_initial_values
from pymcmcstat.ParallelMCMC import check_shape_of_users_initial_values, check_users_initial_values_wrt_limits
from pymcmcstat.ParallelMCMC import get_parameter_features, unpack_mcmc_set
from pymcmcstat.samplers.utilities import is_sample_outside_bounds
from pymcmcstat.settings.ModelParameters import ModelParameters
import test.general_functions as gf
import unittest
from mock import patch
import os
import shutil
import numpy as np

def setup_par_mcmc_basic():
    mcstat = gf.basic_mcmc()
    tmpfolder = gf.generate_temp_folder()
    mcstat.simulation_options.savedir = tmpfolder
    return mcstat, tmpfolder
# -------------------------
class CheckOptionsOutput(unittest.TestCase):
    def test_check_options_output(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.save_to_bin = True
        mcstat.simulation_options.save_to_txt = True
        options = check_options_output(options = mcstat.simulation_options)
        self.assertTrue(options.save_to_bin, msg = 'Expect True')
        self.assertTrue(options.save_to_txt, msg = 'Expect True')
        
    def test_check_options_output_with_bin_false(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.save_to_bin = False
        mcstat.simulation_options.save_to_txt = True
        options = check_options_output(options = mcstat.simulation_options)
        self.assertFalse(options.save_to_bin, msg = 'Expect False')
        self.assertTrue(options.save_to_txt, msg = 'Expect True')
        
    def test_check_options_output_with_txt_false(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.save_to_bin = True
        mcstat.simulation_options.save_to_txt = False
        options = check_options_output(options = mcstat.simulation_options)
        self.assertTrue(options.save_to_bin, msg = 'Expect True')
        self.assertFalse(options.save_to_txt, msg = 'Expect False')
        
    def test_check_options_output_with_both_false(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.save_to_bin = False
        mcstat.simulation_options.save_to_txt = False
        options = check_options_output(options = mcstat.simulation_options)
        self.assertTrue(options.save_to_bin, msg = 'Expect True')
        self.assertFalse(options.save_to_txt, msg = 'Expect False')
        
# -------------------------
class CheckDirectory(unittest.TestCase):
    def test_check_directory(self):
        tmpfolder = gf.generate_temp_folder()
        check_directory(tmpfolder)
        self.assertTrue(os.path.exists(tmpfolder), msg = 'Directory should exist')
        shutil.rmtree(tmpfolder)
        
# -------------------------
class RunSerialSimulation(unittest.TestCase):
    def test_run_serial_simulation(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.nsimu = 100
        mcstat.simulation_options.save_to_bin = False
        mcstat.simulation_options.save_to_txt = False
        simulation_results = run_serial_simulation(mcstat)
        results = simulation_results.results
        self.assertTrue(isinstance(results, dict), msg = 'Expect dictionary return item')
        self.assertEqual(results['nsimu'], 100, msg = 'Expect 100 simulations')
        
# -------------------------
class AssignNumberOfCores(unittest.TestCase):
    @patch('pymcmcstat.ParallelMCMC.cpu_count', return_value = 4)
    def test_assign_number_of_cores(self, mock_cpu_count):
        for nc in range(4):
            num_cores = assign_number_of_cores(num_cores = nc)
            self.assertEqual(num_cores, nc, msg = str('Expect num_cores = {}'.format(nc))) 
            
    @patch('pymcmcstat.ParallelMCMC.cpu_count', return_value = 2)
    def test_assign_number_of_cores_gte_cpu_count(self, mock_cpu_count):
        for nc in range(2, 4):
            num_cores = assign_number_of_cores(num_cores = nc)
            self.assertEqual(num_cores, 2, msg = str('Expect num_cores = {}'.format(2))) 
            
# -------------------------
class CheckShapeOfUsersInitialValues(unittest.TestCase):
    def test_check_shape_of_uiv(self):
        initial_values = np.random.random_sample(size = (4,3))
        num_chain = 4
        npar = 3
        out = check_shape_of_users_initial_values(initial_values = initial_values, num_chain = num_chain, npar = npar)
        self.assertEqual(out[0], num_chain, msg = str('Expect no change to num_chain: {} neq {}'.format(out[0],num_chain)))
        self.assertTrue(np.array_equal(out[1], initial_values), msg = str('Expect no change to initial_values: {} neq {}'.format(out[1],initial_values)))
        
    def test_check_shape_of_uiv_not_same(self):
        initial_values = np.random.random_sample(size = (4,3))
        num_chain = 3
        npar = 2
        out = check_shape_of_users_initial_values(initial_values = initial_values, num_chain = num_chain, npar = npar)
        self.assertEqual(out[0], 4, msg = str('Expect change to num_chain: {} neq {}'.format(out[0],4)))
        self.assertTrue(np.array_equal(out[1], initial_values[:,:npar]), msg = str('Expect change to initial_values: {} neq {}'.format(out[1],initial_values[:,:npar])))
        
# -------------------------
class CheckUsersInitialValuesWRTLimits(unittest.TestCase):
    @patch('pymcmcstat.ParallelMCMC.is_sample_outside_bounds', return_value = False)
    def test_initial_values_wrt_limits_false(self, mock_outside):
        initial_values = np.random.random_sample(size = (4,3))
        low_lim = np.random.random_sample(size = (4,3))
        upp_lim = np.random.random_sample(size = (4,3))
        out = check_users_initial_values_wrt_limits(initial_values = initial_values, low_lim = low_lim, upp_lim = upp_lim)
        self.assertTrue(np.array_equal(out, initial_values), msg = str('Expect no change to initial_values: {} neq {}'.format(out,initial_values)))

    @patch('pymcmcstat.ParallelMCMC.is_sample_outside_bounds', return_value = True)
    def test_initial_values_wrt_limits_true(self, mock_outside):
        initial_values = np.random.random_sample(size = (4,3))
        low_lim = np.random.random_sample(size = (4,3))
        upp_lim = np.random.random_sample(size = (4,3))
        with self.assertRaises(SystemExit, msg = 'Outside bounds not acceptable for initial values'):
            check_users_initial_values_wrt_limits(initial_values = initial_values, low_lim = low_lim, upp_lim = upp_lim)
# -------------------------
class GenerateInitialValues(unittest.TestCase):
    def test_generate_initial_values(self):
        num_chain = 3
        npar = 3
        low_lim = np.zeros([3])
        upp_lim = np.ones([3])
        initial_values = generate_initial_values(num_chain = num_chain, npar = npar, low_lim = low_lim, upp_lim = upp_lim)
        self.assertFalse(is_sample_outside_bounds(initial_values, low_lim, upp_lim), msg = 'Expect initial values to be inside bounds')

# -------------------------
class CheckInitialValues(unittest.TestCase):
    def test_check_initial_values_none(self):
        num_chain = 3
        npar = 3
        low_lim = np.zeros([3])
        upp_lim = np.ones([3])
        initial_values, num_chain_out = check_initial_values(initial_values = None, num_chain = num_chain, npar = npar, low_lim = low_lim, upp_lim = upp_lim)
        self.assertFalse(is_sample_outside_bounds(initial_values, low_lim, upp_lim), msg = 'Expect initial values to be inside bounds')
        self.assertEqual(initial_values.shape, (3,3), msg = 'Expect initial values array shape = (3,3)')
        self.assertEqual(num_chain_out, num_chain, msg = 'Expect num_chain to be unchanged')
        
    def test_check_shape_of_uiv(self):
        initial_values = np.random.random_sample(size = (4,3))
        num_chain = 4
        npar = 3
        low_lim = np.zeros([npar])
        upp_lim = np.ones([npar])
        out = check_initial_values(initial_values = initial_values, num_chain = num_chain, npar = npar, low_lim = low_lim, upp_lim = upp_lim)
        self.assertEqual(out[1], num_chain, msg = str('Expect no change to num_chain: {} neq {}'.format(out[1],num_chain)))
        self.assertTrue(np.array_equal(out[0], initial_values), msg = str('Expect no change to initial_values: {} neq {}'.format(out[0],initial_values)))
        
    def test_check_shape_of_uiv_not_same(self):
        initial_values = np.random.random_sample(size = (4,3))
        num_chain = 3
        npar = 2
        low_lim = np.zeros([npar])
        upp_lim = np.ones([npar])
        out = check_initial_values(initial_values = initial_values, num_chain = num_chain, npar = npar, low_lim = low_lim, upp_lim = upp_lim)
        self.assertEqual(out[1], 4, msg = str('Expect change to num_chain: {} neq {}'.format(out[1],4)))
        self.assertTrue(np.array_equal(out[0], initial_values[:,:npar]), msg = str('Expect change to initial_values: {} neq {}'.format(out[0],initial_values[:,:npar])))

# -------------------------
class GetParameterFeatures(unittest.TestCase):
    def test_get_parameter_features(self):
        MP = ModelParameters()
        MP.add_model_parameter(name = 'a1', theta0=0.1, minimum=0, maximum=1)
        MP.add_model_parameter(name = 'a2', theta0=0.2, minimum=0, maximum=1)
        MP.add_model_parameter(name = 'a3', theta0=0.3, minimum=0, maximum=1)
        npar, low_lim, upp_lim = get_parameter_features(MP.parameters)
        self.assertEqual(npar, 3, msg = 'Expect to find 3 parameters')
        self.assertTrue(np.array_equal(low_lim, np.zeros([3])), msg = str('Expect low_lim all zero: {} neq {}'.format(low_lim,np.zeros([3]))))
        self.assertTrue(np.array_equal(upp_lim, np.ones([3])), msg = str('Expect upp_lim all one: {} neq {}'.format(upp_lim,np.ones([3]))))
        
    def test_get_parameter_features_with_inf_bounds(self):
        MP = ModelParameters()
        MP.add_model_parameter(name = 'a1', theta0=0.1, minimum=-np.inf, maximum=1)
        MP.add_model_parameter(name = 'a2', theta0=0.2, minimum=0, maximum=1)
        MP.add_model_parameter(name = 'a3', theta0=0.3, minimum=0, maximum=np.inf)
        npar, low_lim, upp_lim = get_parameter_features(MP.parameters)
        expect_low_lim = np.array([0.1 - 100*0.1, 0, 0])
        expect_upp_lim = np.array([1, 1, 0.3 + 100*0.3])
        self.assertEqual(npar, 3, msg = 'Expect to find 3 parameters')
        self.assertTrue(np.array_equal(low_lim, expect_low_lim), msg = str('Expect low_lim finite: {} neq {}'.format(low_lim,expect_low_lim)))
        self.assertTrue(np.array_equal(upp_lim, expect_upp_lim), msg = str('Expect upp_lim finite: {} neq {}'.format(upp_lim,expect_upp_lim)))
        
# -------------------------
class UnpackMCMCSet(unittest.TestCase):
    def test_unpack_mcmc_set(self):
        mcstat = gf.basic_mcmc()
        data, options, model, parameters = unpack_mcmc_set(mcset = mcstat)
        self.assertEqual(data, mcstat.data, msg = 'Expect structures to match')
        self.assertEqual(options, mcstat.simulation_options, msg = 'Expect structures to match')
        self.assertEqual(model, mcstat.model_settings, msg = 'Expect structures to match')
        self.assertEqual(parameters, mcstat.parameters, msg = 'Expect structures to match')

# -------------------------
class ParallelMCMCInit(unittest.TestCase):
    def test_par_mc_init(self):
        PMC = ParallelMCMC()
        check_these = ['setup_parallel_simulation', 'run_parallel_simulation', 'display_individual_chain_statistics', 'description']
        for ct in check_these:
            self.assertTrue(hasattr(PMC, ct), msg = str('Expect class to have attribute: {}'.format(ct)))
            
# -------------------------
class SetupParallelMCMC(unittest.TestCase):
    def test_setup_parmc(self):
        PMC = ParallelMCMC()
        mcstat, tmpfolder = setup_par_mcmc_basic()
        PMC.setup_parallel_simulation(mcset = mcstat)
        self.assertEqual(PMC.num_cores, 1, msg = 'Default cores = 1')
        self.assertEqual(PMC.num_chain, 1, msg = 'Default chains = 1')
        self.assertTrue(isinstance(PMC.parmc, list), msg = 'Expect list')
        self.assertEqual(len(PMC.parmc), 1, msg = 'Expect length of 1')
        shutil.rmtree(tmpfolder)
        
    def test_setup_parmc_with_initial_values(self):
        PMC = ParallelMCMC()
        mcstat, tmpfolder = setup_par_mcmc_basic()
        initial_values = np.array([[0.5, 0.5, 0.5]])
        PMC.setup_parallel_simulation(mcset = mcstat, initial_values = initial_values)
        self.assertEqual(PMC.num_cores, 1, msg = 'Default cores = 1')
        self.assertEqual(PMC.num_chain, 1, msg = 'Default chains = 1')
        self.assertTrue(isinstance(PMC.parmc, list), msg = 'Expect list')
        self.assertEqual(len(PMC.parmc), 1, msg = 'Expect length of 1')
        self.assertTrue(np.array_equal(PMC.initial_values, initial_values), msg = str('Expect arrays to match: {} neq {}'.format(PMC.initial_values, initial_values)))
        shutil.rmtree(tmpfolder)
        
    @patch('pymcmcstat.ParallelMCMC.cpu_count', return_value = 2)
    def test_setup_parmc_with_initial_values_num_cores_chain_2(self, mock_cpu_count):
        PMC = ParallelMCMC()
        mcstat, tmpfolder = setup_par_mcmc_basic()
        initial_values = np.array([[0.5, 0.5, 0.5], [0.5, 0.5, 0.5]])
        PMC.setup_parallel_simulation(mcset = mcstat, initial_values = initial_values, num_cores = 2, num_chain = 2)
        self.assertEqual(PMC.num_cores, 2, msg = 'Default cores = 2')
        self.assertEqual(PMC.num_chain, 2, msg = 'Default chains = 2')
        self.assertTrue(isinstance(PMC.parmc, list), msg = 'Expect list')
        self.assertEqual(len(PMC.parmc), 2, msg = 'Expect length of 2')
        self.assertTrue(np.array_equal(PMC.initial_values, initial_values), msg = str('Expect arrays to match: {} neq {}'.format(PMC.initial_values, initial_values)))
        shutil.rmtree(tmpfolder)
# -------------------------
class RunParallelMCMC(unittest.TestCase):
    def test_setup_parmc(self):
        PMC = ParallelMCMC()
        mcstat, tmpfolder = setup_par_mcmc_basic()
        PMC.setup_parallel_simulation(mcset = mcstat)
        PMC.run_parallel_simulation()
        self.assertEqual(PMC.num_cores, 1, msg = 'Default cores = 1')
        self.assertEqual(PMC.num_chain, 1, msg = 'Default chains = 1')
        self.assertTrue(isinstance(PMC.parmc, list), msg = 'Expect list')
        self.assertEqual(len(PMC.parmc), 1, msg = 'Expect length of 1')
        self.assertTrue(hasattr(PMC.parmc[0], 'simulation_results'), msg = 'Expect results added')
        self.assertEqual(PMC.parmc[0].simulation_results.results['nsimu'], 5000, msg = 'Expect nsimu = 5000')
        shutil.rmtree(tmpfolder)


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 10:11:40 2018

@author: prmiles
"""
# import required packages
#import numpy as np
from .Metropolis import Metropolis
from .Adaptation import Adaptation
from .DelayedRejection import DelayedRejection

class SamplingMethods:
    '''
    Metropolis sampling methods.
    
    Attributes:
        * :class:`~.Metropolis`
        * :class:`~.DelayedRejection`
        * :class:`~.Adaptation`
    '''
    def __init__(self):
        self.metropolis = Metropolis()
        self.delayed_rejection = DelayedRejection()
        self.adaptation = Adaptation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 27 12:08:17 2018

@author: prmiles
"""
from pymcmcstat.samplers.utilities import sample_candidate_from_gaussian_proposal
from pymcmcstat.samplers.utilities import is_sample_outside_bounds
from pymcmcstat.samplers.utilities import acceptance_test
from pymcmcstat.samplers.utilities import set_outside_bounds
from pymcmcstat.structures.ParameterSet import ParameterSet
import unittest
from mock import patch
import numpy as np

# --------------------------
class SampleCandidate(unittest.TestCase):
    @patch('numpy.random.randn')
    def test_sample_candidate(self, mock_simple_func):
        mock_simple_func.return_value = np.array([0.1, 0.2])
        oldpar = np.array([0.1, 0.4])
        R = np.array([[0.4, 0.2],[0, 0.3]])
        newpar, npar_sample_from_normal = sample_candidate_from_gaussian_proposal(npar = 2, oldpar = oldpar, R = R)
        self.assertEqual(newpar.size, 2, msg = 'Size of parameter array is 2')
        self.assertEqual(npar_sample_from_normal.size, 2, msg = 'Size of sample is 2')
        self.assertTrue(np.array_equal(newpar, (oldpar + np.dot(np.array([0.1, 0.2]), R)).reshape(2)), msg = 'Arrays should match')

# --------------------------
class OutsideBounds(unittest.TestCase):
    def test_outsidebounds_p1_below(self):
        self.assertTrue(is_sample_outside_bounds(theta = np.array([-1.0, 1.5]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p1 below')
        self.assertTrue(is_sample_outside_bounds(theta = np.array([1.0, 0.5]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p2 below')
        self.assertTrue(is_sample_outside_bounds(theta = np.array([1.5, 1.5]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p1 above')
        self.assertTrue(is_sample_outside_bounds(theta = np.array([1.0, 2.5]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p2 above')
        self.assertFalse(is_sample_outside_bounds(theta = np.array([0, 1.5]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p1 on low lim')
        self.assertFalse(is_sample_outside_bounds(theta = np.array([0.5, 1]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p2 on low lim')
        self.assertFalse(is_sample_outside_bounds(theta = np.array([1, 1.5]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p1 on upp lim')
        self.assertFalse(is_sample_outside_bounds(theta = np.array([0.5, 2]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'p2 on upp lim')
        self.assertFalse(is_sample_outside_bounds(theta = np.array([0.5, 1.5]), lower_limits = np.array([0,1]), upper_limits=np.array([1,2])), msg = 'all outside')
        
# --------------------------        
class Acceptance(unittest.TestCase):
    
    def test_accept_alpha_gt_1(self):
        self.assertEqual(acceptance_test(alpha = 1.1), 1, msg = 'accept alpha >= 1')
        self.assertEqual(acceptance_test(alpha = 1), 1, msg = 'accept alpha >= 1')
        self.assertEqual(acceptance_test(alpha = -0.1), 0, msg = 'Reject alpha <= 0')
        self.assertEqual(acceptance_test(alpha = 0), 0, msg = 'Reject alpha <= 0')
        np.random.seed(0)
        self.assertEqual(acceptance_test(alpha = 0.4), 0, msg = 'Reject alpha < u (0.4 < 0.5488135)')
        np.random.seed(0)
        self.assertEqual(acceptance_test(alpha = 0.6), 1, msg = 'Accept alpha > u (0.6 > 0.5488135)')
        
# --------------------------        
class SetOutsideBounds(unittest.TestCase):
    def test_set_outsidebounds(self):
        next_set = ParameterSet()
        next_set, outbound = set_outside_bounds(next_set = next_set)
        self.assertEqual(next_set.alpha, 0, msg = 'next_set.alpha should be 0')
        self.assertEqual(next_set.prior, 0, msg = 'next_set.prior should be 0')
        self.assertEqual(next_set.ss, np.inf, msg = 'next_set.ss should be np.inf')
        self.assertEqual(outbound, 1, msg = 'outbound should be 1')  


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 07:55:46 2018

Description: Support methods for initializing and updating the covariance matrix.
Additional routines associated with Cholesky Decomposition.

@author: prmiles
"""

# import required packages
import numpy as np
import math

class CovarianceProcedures:
    '''
    Covariance matrix variables and methods.

    Attributes:
        * :meth:`~display_covariance_settings`
        * :meth:`~setup_covariance_matrix`
    '''
    def __init__(self):
        self.description = 'Covariance Variables and Methods'

    def _initialize_covariance_settings(self, parameters, options):
        '''
        Initialize covariance settings.

        Args:
            * **parameters** (:class:`~.ModelParameters`): MCMC model parameters
            * **options** (:class:`~.SimulationOptions`): MCMC simulation options
        '''
        self._qcov = None
        self._qcov_scale = None
        self._R = None
        self._qcov_original = None
        self._invR = None
        self._iacce = None
        self._covchain = None
        self._meanchain = None
        self._last_index_since_adaptation = 0
        
        self._wsum = options.initqcovn
        
        # define noadaptind as a boolean - inputted as list of index values not updated
        self.__setup_no_adapt_index(noadaptind = options.noadaptind, parind = parameters._parind)
        
        # ----------------
        # setup covariance matrix
        self.setup_covariance_matrix(options.qcov, parameters._thetasigma, parameters._initial_value)
            
        # ----------------
        # check adascale
        self.__check_adascale(options.adascale, parameters.npar)
        
        # ----------------
        # setup R matrix (R used to update in metropolis)
#        print('qcov.shape = {}'.format(self._qcov.shape))
        self.__setup_R_matrix(parameters._parind)
            
        # ----------------
        # setup RDR matrix (RDR used in DR)
        if options.method == 'dram' or options.method == 'dr':
            self._invR = []
            self.__setup_RDR_matrix(npar = parameters.npar,
                    drscale = options.drscale, ntry = options.ntry,
                    RDR = options.RDR)
            
    def _update_covariance_from_adaptation(self, R, covchain, meanchain, wsum,
                                          last_index_since_adaptation, iiadapt):
        '''
        Update covariance from adaptation algorithm.

        Args:
            * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
            * **covchain** (:class:`~numpy.ndarray`): Covariance matrix history.
            * **meanchain** (:class:`~numpy.ndarray`): Current mean chain values.
            * **wsum** (:class:`~numpy.ndarray`): Weights
            * **last_index_since_adaptation** (:py:class:`int`): Last index since adaptation occured.
            * **iiadapt** (:py:class:`int`): Adaptation counter.
        '''
        self._R = R
        self._covchain = covchain
        self._meanchain = meanchain
        self._wsum = wsum
        self._last_index_since_adaptation = last_index_since_adaptation
        self._iiadapt = iiadapt
        
    def _update_covariance_for_delayed_rejection_from_adaptation(self, RDR = None, invR = None):
        '''
        Update covariance variables for delayed rejection based on adaptation.

        Args:
            * **RDR** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix based on DR.
            * **invR** (:class:`~numpy.ndarray`): Inverse of Cholesky decomposition matrix.
        '''
        self._RDR = RDR
        self._invR = invR
        
    def _update_covariance_settings(self, parameter_set):
        '''
        Update covariance settings based on parameter set

        Args:
            * **parameter_set** (:class:`~numpy.ndarray`): Mean parameter values
        '''
        if self._wsum is not None:
            self._covchain = self._qcov
            self._meanchain = parameter_set
        
    def setup_covariance_matrix(self, qcov, thetasig, value):
        '''
        Initialize covariance matrix.

        If no proposal covariance matrix is provided, then the default is generated
        by squaring 5% of the initial value.  This yields a diagonal covariance matrix.

        .. math::

            V = diag([(0.05\\theta_i)^2])

        If the initial value was one, this would lead to zero variance.  In those
        instances the variance is set equal to :code:`qcov[qcov==0] = 1.0`.

        Args:
            * **qcov** (:class:`~numpy.ndarray`): Parameter covariance matrix.
            * **thetasig** (:class:`~numpy.ndarray`): Prior variance.
            * **value** (:class:`~numpy.ndarray`): Current parameter value.
        '''
        # check qcov
        if qcov is None: # i.e., qcov is None (not defined)
            qcov = thetasig**2 # variance
            ii1 = np.isinf(qcov)
            ii2 = np.isnan(qcov)
            ii = ii1 + ii2
            qcov[ii] = (np.abs(value[ii])*0.05)**2 # default is 5% stdev
            qcov[qcov==0] = 1.0 # if initial value was zero, use 1 as stdev
            qcov = np.diagflat(qcov) # create covariance matrix
    
        self._qcov = np.atleast_2d(qcov[:])
        
    def __check_adascale(self, adascale, npar):
        # check adascale
        if adascale is None or adascale <= 0:
            qcov_scale = 2.4*(math.sqrt(npar)**(-1)) # scale factor in R
        else:
            qcov_scale = adascale
        
        self._qcov_scale = qcov_scale
    
    def __setup_R_matrix(self, parind):
        cm, cn = self._qcov.shape # number of rows, number of columns
        if min([cm, cn]) == 1: # qcov contains variances!
            self.__setup_R_based_on_variances(parind)
        else: # qcov has covariance matrix in it
            self.__setup_R_based_on_covariance_matrix(parind)
    
    def __setup_R_based_on_variances(self, parind):
        qcov = np.copy(self._qcov)
        qcov = np.diagflat(qcov)
        self._R = np.sqrt(qcov[np.ix_(parind,parind)])
        self._qcovorig = np.diagflat(self._qcov[:]) # save original qcov
        self._qcov = qcov[np.ix_(parind,parind)]
        
    def __setup_R_based_on_covariance_matrix(self, parind):
        self._qcovorig = np.copy(self._qcov) # save qcov
        self._qcov = self._qcov[np.ix_(parind,parind)] # this operation in matlab maintains matrix (debug)
        if self._qcov.size == 1:
            self._R = np.sqrt(self._qcov)
        else:
            self._R = np.linalg.cholesky(self._qcov) # cholesky decomposition
        self._R = self._R.transpose() # matches output of matlab function
            
    def __setup_RDR_matrix(self, npar, drscale, ntry, RDR):
        # if not empty
        if RDR is None: # check implementation
            RDR = [] # initialize list
            RDR.append(self._R)
            self._invR.append(np.linalg.solve(self._R, np.eye(npar)))
            for ii in range(1,ntry):
                RDR.append(RDR[ii-1]*(drscale[min(ii,len(drscale))-1]**(-1)))
                self._invR.append(np.linalg.solve(RDR[ii],np.eye(npar)))
        else: # DR strategy: just scale R's down by DR_scale
            for ii in range(ntry):
                self._invR.append(np.linalg.solve(RDR[ii], np.eye(npar)))
                    
            self._R = RDR[0]
        
        self._RDR = RDR
        
    def __setup_no_adapt_index(self, noadaptind, parind):
        # define noadaptind as a boolean - inputted as list of index values not updated
        no_adapt_index = np.zeros([len(parind)],dtype=bool)
        if len(noadaptind) == 0:
            no_adapt_index = np.zeros([len(parind)],dtype=bool)
        else:
            c = list(set.intersection(set(noadaptind), set(parind)))
            no_adapt_index[c] = np.ones([1], dtype = bool)
    
        self._no_adapt_index = no_adapt_index
        
    def display_covariance_settings(self, print_these = None):
        '''
        Display subset of the covariance settings.

        Args:
            * **print_these** (:py:class:`list`): List of strings corresponding to keywords.  Default below.

        ::

            print_these = ['qcov', 'R', 'RDR', 'invR', 'last_index_since_adaptation', 'covchain']
        '''
        if print_these is None:
            print_these = ['qcov', 'R', 'RDR', 'invR', 'last_index_since_adaptation', 'covchain']
        print('covariance:')
        for ptii in print_these:
            print('\t{} = {}'.format(ptii, getattr(self, str('_{}'.format(ptii)))))
        return print_these


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 31 12:54:16 2018

@author: prmiles
"""

# import required packages
from __future__ import division
import math
import matplotlib.pyplot as plt
from pylab import hist
from .utilities import generate_names, setup_plot_features, make_x_grid

import warnings

try:
    from statsmodels.nonparametric.kernel_density import KDEMultivariate
except ImportError as e:
    warnings.warn(str("Exception raised importing statsmodels.nonparametric.kernel_density - plot_density_panel will not work. {}".format(e)))

# --------------------------------------------
def plot_density_panel(chains, names = None, hist_on = False, figsizeinches = None):
    '''
    Plot marginal posterior densities

    Args:
        * **chains** (:class:`~numpy.ndarray`): Sampling chain for each parameter
        * **names** (:py:class:`list`): List of strings - name of each parameter
        * **hist_on** (:py:class:`bool`): Flag to include histogram on density plot
        * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height]
    '''
    nsimu, nparam = chains.shape # number of rows, number of columns
    ns1, ns2, names, figsizeinches = setup_plot_features(nparam = nparam, names = names, figsizeinches = figsizeinches)

    f = plt.figure(dpi=100, figsize=(figsizeinches)) # initialize figure
    for ii in range(nparam):
        # define chain
        chain = chains[:,ii].reshape(nsimu,1) # check indexing
        
        # define x grid
        chain_grid = make_x_grid(chain)
        
        # Compute kernel density estimate
        kde = KDEMultivariate(chain, bw = 'normal_reference', var_type = 'c')

        # plot density on subplot
        plt.subplot(ns1,ns2,ii+1)
             
        if hist_on == True: # include histograms
            hist(chain, normed=True)
            
        plt.plot(chain_grid, kde.pdf(chain_grid), 'k')
        # format figure
        plt.xlabel(names[ii])
        plt.ylabel(str('$\pi$({}$|M^{}$)'.format(names[ii], '{data}')))
        plt.tight_layout(rect=[0, 0.03, 1, 0.95],h_pad=1.0) # adjust spacing

    return f

# --------------------------------------------
def plot_histogram_panel(chains, names = None, figsizeinches = None):
    """
    Plot histogram from each parameter's sampling history

    Args:
        * **chains** (:class:`~numpy.ndarray`): Sampling chain for each parameter
        * **names** (:py:class:`list`): List of strings - name of each parameter
        * **hist_on** (:py:class:`bool`): Flag to include histogram on density plot
        * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height]
    """
    nsimu, nparam = chains.shape # number of rows, number of columns
    ns1, ns2, names, figsizeinches = setup_plot_features(nparam = nparam, names = names, figsizeinches = figsizeinches)
        
    f = plt.figure(dpi=100, figsize=(figsizeinches)) # initialize figure
    for ii in range(nparam):
        # define chain
        chain = chains[:,ii].reshape(nsimu,1) # check indexing
        
        # plot density on subplot
        ax = plt.subplot(ns1,ns2,ii+1)
        hist(chain, normed=True)
        # format figure
        plt.xlabel(names[ii])
        ax.set_yticklabels([])
        plt.tight_layout(rect=[0, 0.03, 1, 0.95],h_pad=1.0) # adjust spacing
        
    return f
        
# --------------------------------------------
def plot_chain_panel(chains, names = None, figsizeinches = None, maxpoints = 500):
    """
    Plot sampling chain for each parameter

    Args:
        * **chains** (:class:`~numpy.ndarray`): Sampling chain for each parameter
        * **names** (:py:class:`list`): List of strings - name of each parameter
        * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height]
        * **maxpoints** (:py:class:`int`): Max number of display points - keeps scatter plot from becoming overcrowded
    """
    nsimu, nparam = chains.shape # number of rows, number of columns
    ns1, ns2, names, figsizeinches = setup_plot_features(nparam = nparam, names = names, figsizeinches = figsizeinches)
    
    skip = 1
    if nsimu > maxpoints:
        skip = int(math.floor(nsimu/maxpoints))
    
    f = plt.figure(dpi=100, figsize=(figsizeinches)) # initialize figure
    for ii in range(nparam):
        # define chain
        chain = chains[:,ii].reshape(nsimu,1) # check indexing
        
        # plot chain on subplot
        plt.subplot(ns1,ns2,ii+1)
        plt.plot(range(0,nsimu,skip), chain[range(0,nsimu,skip),0], '.b')
        # format figure
        plt.xlabel('Iteration')
        plt.ylabel(str('{}'.format(names[ii])))
        if ii+1 <= ns1*ns2 - ns2:
            plt.xlabel('')
            
        plt.tight_layout(rect=[0, 0.03, 1, 0.95],h_pad=1.0) # adjust spacing
        
    return f
        
# --------------------------------------------
def plot_pairwise_correlation_panel(chains, names = None, figsizeinches = None, skip = 1):
    """
    Plot pairwise correlation for each parameter

    Args:
        * **chains** (:class:`~numpy.ndarray`): Sampling chain for each parameter
        * **names** (:py:class:`list`): List of strings - name of each parameter
        * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height]
        * **skip** (:py:class:`int`): Indicates step size to be used when plotting elements from the chain
    """
    nsimu, nparam = chains.shape # number of rows, number of columns
    
    inds = range(0,nsimu,skip)
    
    names = generate_names(nparam = nparam, names = names)
        
    if figsizeinches is None:
        figsizeinches = [7,5]
        
    f = plt.figure(dpi=100, figsize=(figsizeinches)) # initialize figure
    for jj in range(2,nparam+1):
        for ii in range(1,jj):
            chain1 = chains[inds,ii-1]
            chain1 = chain1.reshape(nsimu,1)
            chain2 = chains[inds,jj-1]
            chain2 = chain2.reshape(nsimu,1)
            
            # plot density on subplot
            ax = plt.subplot(nparam-1,nparam-1,(jj-2)*(nparam-1)+ii)
            plt.plot(chain1, chain2, '.b')
            
            # format figure
            if jj != nparam: # rm xticks
                ax.set_xticklabels([])
            if ii != 1: # rm yticks
                ax.set_yticklabels([])
            if ii == 1: # add ylabels
                plt.ylabel(str('{}'.format(names[jj-1])))
            if ii == jj - 1:
                if nparam == 2: # add xlabels
                    plt.xlabel(str('{}'.format(names[ii-1])))
                else: # add title
                    plt.title(str('{}'.format(names[ii-1])))
         
    # adjust figure margins
    plt.tight_layout(rect=[0, 0.03, 1, 0.95],h_pad=1.0) # adjust spacing
    
    return f
 
# --------------------------------------------
def plot_chain_metrics(chain, name = None, figsizeinches = None):
    '''
    Plot chain metrics for individual chain

    - Scatter plot of chain
    - Histogram of chain
    
    Args:
        * **chains** (:class:`~numpy.ndarray`): Sampling chain for specific parameter
        * **names** (:py:class:`str`): Name of each parameter
        * **figsizeinches** (:py:class:`list`): Specify figure size in inches [Width, Height]
    '''
    name = generate_names(nparam = 1, names = name)
    
    if figsizeinches is None:
        figsizeinches = [7,5]
        
    f = plt.figure(dpi=100, figsize=(figsizeinches)) # initialize figure
    plt.suptitle('Chain metrics for {}'.format(name), fontsize='12')
    plt.subplot(2,1,1)
    plt.plot(range(0,len(chain)),chain, marker='.')
    # format figure
    plt.xlabel('Iterations')
    ystr = str('{}-chain'.format(name))
    plt.ylabel(ystr)
    # Add histogram
    plt.subplot(2,1,2)
    hist(chain)
    # format figure
    plt.xlabel(name)
    plt.ylabel(str('Histogram of {}-chain'.format(name)))
    plt.tight_layout(rect=[0, 0.03, 1, 0.95],h_pad=1.0) # adjust spacing
    return f
    
class Plot:
    '''
    Plotting routines for analyzing sampling chains from MCMC process.

    Attributes:
        - :meth:`~plot_density_panel`
        - :meth:`~plot_chain_panel`
        - :meth:`~plot_pairwise_correlation_panel`
        - :meth:`~plot_histogram_panel`
        - :meth:`~plot_chain_metrics`
    '''
    def __init__(self):
        self.plot_density_panel = plot_density_panel
        self.plot_chain_panel = plot_chain_panel
        self.plot_pairwise_correlation_panel = plot_pairwise_correlation_panel
        self.plot_histogram_panel = plot_histogram_panel
        self.plot_chain_metrics = plot_chain_metrics


from setuptools import setup, find_packages
import codecs

def read(fname):
    with codecs.open(fname, 'r', 'latin') as f:
        return f.read()
    
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with codecs.open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
    
# read in version number
version_dummy = {}
exec(read('pymcmcstat/__version__.py'), version_dummy)
__version__ = version_dummy['__version__']
del version_dummy

setup(
    name='pymcmcstat',
    version=__version__,
    description=('A library to perform MCMC simulations using DRAM'),
    long_description=long_description,
    url='https://github.com/prmiles/pymcmcstat',
    download_url='https://github.com/prmiles/pymcmcstat',
    author='Paul Miles',
    author_email='prmiles@ncsu.edu',
    license='MIT',
    package_dir={'pymcmcstat': 'pymcmcstat'},
    packages=find_packages(),
    zip_safe=False,
    install_requires=['numpy>=1.14', 'scipy>=1.0', 'matplotlib>=2.2.0', 'h5py>=2.7.0', 'statsmodels>=0.9.0'],
    extras_require = {'docs':['sphinx'], 'plotting':['matplotlib', 'plotly'],},
    classifiers=['License :: OSI Approved :: MIT License',
                   'Natural Language :: English',
                   'Operating System :: MacOS :: MacOS X',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python :: 3.6',
                   'Framework :: IPython',
                   'Intended Audience :: Science/Research',
                   'Topic :: Scientific/Engineering',
                   ]
)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 11:14:11 2018

@author: prmiles
"""
# import required packages
import numpy as np
import math
from ..utilities.general import message

# --------------------------------------------
class Adaptation:
    '''
    Adaptive Metropolis (AM) algorithm based on :cite:`haario2001adaptive`.

    Attributes:
        * :meth:`~run_adaptation`
    '''
    def __init__(self):
        self.label = 'Covariance Variables and Methods'
        self.qcov = None
        self.qcov_scale = None
        self.R = None
        self.qcov_original = None
        self.invR = None
        self.iacce = None
        self.covchain = None
        self.meanchain = None
        self.last_index_since_adaptation = 0

    # -------------------------------------------
    def run_adaptation(self, covariance, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha):
        '''
        Run adaptation step

        Args:
            * **covariance** (:class:`~.CovarianceProcedures`): Covariance methods and variables
            * **options** (:class:`~.SimulationOptions`): Options for MCMC simulation
            * **isimu** (:py:class:`int`): Simulation counter
            * **iiadapt** (:py:class:`int`): Adaptation counter
            * **rejected** (:py:class:`dict`): Rejection counter
            * **chain** (:class:`~numpy.ndarray`): Sampling chain
            * **chainind** (:py:class:`ind`): Relative point in chain
            * **u** (:class:`~numpy.ndarray`): Latest random sample points
            * **npar** (:py:class:`int`): Number of parameters being sampled
            * **alpha** (:py:class:`float`): Latest Likelihood evaluation

        Returns:
            * **covariance** (:class:`~.CovarianceProcedures`): Updated covariance object
        '''
        # unpack options
        burnintime, burnin_scale, ntry, drscale, alphatarget, etaparam, qcov_adjust, doram, verbosity = unpack_simulation_options(options = options)
        
        # unpack covariance
        last_index_since_adaptation, R, oldcovchain, oldmeanchain, oldwsum, no_adapt_index, qcov_scale, qcov = unpack_covariance_settings(covariance = covariance)
        
        if isimu < burnintime:
            R = below_burnin_threshold(rejected = rejected, iiadapt = iiadapt, R = R, burnin_scale = burnin_scale, verbosity = verbosity)
            covchain = oldcovchain
            meanchain = oldmeanchain
            wsum = oldwsum
            RDR = None
            invR = None
        else:
            message(verbosity, 2, str('i:{} adapting ({}, {}, {})'.format(
                    isimu, rejected['total']*(isimu**(-1))*100, rejected['in_adaptation_interval']*(iiadapt**(-1))*100,
                    rejected['outside_bounds']*(isimu**(-1))*100)))
    
            # UPDATE COVARIANCE MATRIX - CHOLESKY, MEAN, SUM
            covchain, meanchain, wsum = update_covariance_mean_sum(
                    chain[last_index_since_adaptation:chainind,:], np.ones(1), oldcovchain, oldmeanchain, oldwsum)
                    
            last_index_since_adaptation = isimu
                    
            # ram
            if doram:
                upcov = update_cov_via_ram(u = u, isimu = isimu, etaparam=etaparam, npar = npar, alphatarget = alphatarget, alpha = alpha, R = R)
            else:
                upcov = update_cov_from_covchain(covchain = covchain, qcov = qcov, no_adapt_index = no_adapt_index)

            # check if singular covariance matrix
            R = check_for_singular_cov_matrix(upcov = upcov, R = R, npar = npar, qcov_adjust = qcov_adjust, qcov_scale = qcov_scale, rejected = rejected, iiadapt = iiadapt, verbosity = verbosity)
            
            # update dram covariance matrix
            RDR, invR = update_delayed_rejection(R = R, npar = npar, ntry = ntry, drscale = drscale)

        covariance._update_covariance_from_adaptation(R, covchain, meanchain, wsum,
                                          last_index_since_adaptation, iiadapt)

        covariance._update_covariance_for_delayed_rejection_from_adaptation(RDR = RDR, invR = invR)
        
        return covariance
# --------------------------------------------
def unpack_simulation_options(options):
    '''
    Unpack simulation options

    Args:
        * **options** (:class:`~.SimulationOptions`): Options for MCMC simulation

    Returns:
        * **burnintime** (:py:class:`int`):
        * **burnin_scale** (:py:class:`float`): Scale for burnin.
        * **ntry** (:py:class:`int`): Number of tries to take before rejection. Default is method dependent.
        * **drscale** (:class:`~numpy.ndarray`): Reduced scale for sampling in DR algorithm. Default is [5,4,3].
        * **alphatarget** (:py:class:`float`): Acceptance ratio target.
        * **etaparam** (:py:class:`float`):
        * **qcov_adjust** (:py:class:`float`): Adjustment scale for covariance matrix.
        * **doram** (:py:class:`int`): Flag to perform :code:`'ram'` algorithm (Obsolete).
        * **verbosity** (:py:class:`int`): Verbosity of display output.
    '''
    burnintime = options.burnintime
    burnin_scale = options.burnin_scale
    ntry = options.ntry
    drscale = options.drscale
    alphatarget = options.alphatarget
    etaparam = options.etaparam
    qcov_adjust = options.qcov_adjust
    doram = options.doram
    verbosity = options.verbosity
    return burnintime, burnin_scale, ntry, drscale, alphatarget, etaparam, qcov_adjust, doram, verbosity
# --------------------------------------------
def unpack_covariance_settings(covariance):
    '''
    Unpack covariance settings

    Args:
        * **covariance** (:class:`~.CovarianceProcedures`): Covariance methods and variables

    Returns:
        * **last_index_since_adaptation** (:py:class:`int`): Last index since adaptation occured.
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
        * **oldcovchain** (:class:`~numpy.ndarray`): Covariance matrix history.
        * **oldmeanchain** (:class:`~numpy.ndarray`): Current mean chain values.
        * **oldwsum** (:class:`~numpy.ndarray`): Weights
        * **no_adapt_index** (:class:`numpy.ndarray`): Indices of parameters not being adapted.
        * **qcov_scale** (:py:class:`float`): Scale parameter
        * **qcov** (:class:`~numpy.ndarray`): Covariance matrix
    '''
    # unpack covariance
    last_index_since_adaptation = covariance._last_index_since_adaptation
    R = covariance._R
    oldcovchain = covariance._covchain
    oldmeanchain = covariance._meanchain
    oldwsum = covariance._wsum
    no_adapt_index = covariance._no_adapt_index
    
    qcov_scale = covariance._qcov_scale
    qcov = covariance._qcov
    return last_index_since_adaptation, R, oldcovchain, oldmeanchain, oldwsum, no_adapt_index, qcov_scale, qcov
# --------------------------------------------
def below_burnin_threshold(rejected, iiadapt, R, burnin_scale, verbosity):
    '''
    Update Cholesky Matrix using below burnin thershold

    Args:
        * **rejected** (:py:class:`dict`): Rejection counters.
        * **iiadapt** (:py:class:`int`): Adaptation counter
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
        * **burnin_scale** (:py:class:`float`): Scale for burnin.
        * **verbosity** (:py:class:`int`): Verbosity of display output.

    Returns:
        * **R** (:class:`~numpy.ndarray`): Scaled Cholesky matrix.
    '''
    # during burnin no adaptation, just scaling down
    if rejected['in_adaptation_interval']*(iiadapt**(-1)) > 0.95:
        message(verbosity, 2, str(' (burnin/down) {}'.format(rejected['in_adaptation_interval']*(iiadapt**(-1))*100)))
        R = R*(burnin_scale**(-1))
    elif rejected['in_adaptation_interval']*(iiadapt**(-1)) < 0.05:
        message(verbosity, 2, str(' (burnin/up) {}'.format(
                rejected['in_adaptation_interval']*(iiadapt**(-1))*100)))
        R = R*burnin_scale
    return R
# --------------------------------------------
def update_covariance_mean_sum(x, w, oldcov, oldmean, oldwsum, oldR = None):
    '''
    Update covariance chain, local mean, local sum

    Args:
        * **x** (:class:`~numpy.ndarray`): Chain segment
        * **w** (:class:`~numpy.ndarray`): Weights
        * **oldcov** (:class:`~numpy.ndarray` or `None`): Previous covariance matrix
        * **oldmean** (:class:`~numpy.ndarray`): Previous mean chain values
        * **oldwsum** (:class:`~numpy.ndarray`): Previous weighted sum
        * **oldR** (:class:`~numpy.ndarray`): Previous Cholesky decomposition matrix

    Returns:
        * **xcov** (:class:`~numpy.ndarray`): Updated covariance matrix
        * **xmean** (:class:`~numpy.ndarray`): Updated mean chain values
        * **wsum** (:class:`~numpy.ndarray`): Updated weighted sum
    '''
    n, p = x.shape
    
    if n == 0 or p == 0: # nothing to update with
        return oldcov, oldmean, oldwsum
    
    w, R = setup_w_R(w = w, oldR = oldR, n = n)
           
    if oldcov is None:
        xcov, xmean, wsum = initialize_covariance_mean_sum(x, w)
                        
    else:
        for ii in range(0,n):
            xi = x[ii,:]
            wsum = w[ii]
            xmean = oldmean + wsum*((wsum+oldwsum)**(-1.0))*(xi - oldmean)
            
            if R is not None:
                Rin, xin = setup_cholupdate(R = R, oldwsum = oldwsum, wsum = wsum, oldmean = oldmean, xi = xi)
                R = cholupdate(Rin, xin)
        
            xcov = (((oldwsum-1)*((wsum + oldwsum - 1)**(-1)))*oldcov
                    + (wsum*oldwsum*((wsum+oldwsum-1)**(-1)))*((wsum
                           + oldwsum)**(-1))*(np.dot((xi-oldmean).reshape(p,1),(xi-oldmean).reshape(1,p))))
        
            wsum = wsum + oldwsum
            oldcov = xcov
            oldmean = xmean
            oldwsum = wsum
   
    return xcov, xmean, wsum
# --------------------------------------------
def setup_w_R(w, oldR, n):
    '''
    Setup weights and Cholesky matrix

    Args:
        * **x** (:class:`~numpy.ndarray`): Chain segment
        * **w** (:class:`~numpy.ndarray`): Weights
        * **oldcov** (:class:`~numpy.ndarray` or `None`): Previous covariance matrix
        * **oldmean** (:class:`~numpy.ndarray`): Previous mean chain values
        * **oldwsum** (:class:`~numpy.ndarray`): Previous weighted sum
        * **oldR** (:class:`~numpy.ndarray`): Previous Cholesky decomposition matrix

    Returns:
        * **w** (:class:`~numpy.ndarray`): Weights
        * **R** (:class:`~numpy.ndarray`): Previous Cholesky decomposition matrix
    '''
    if w is None:
        w = np.ones(1)
        
    if len(w) == 1:
        w = np.ones(n)*w
        
    if oldR is None:
        R = None
    else:
        R = oldR
    return w, R
# --------------------------------------------
def initialize_covariance_mean_sum(x, w):
    '''
    Initialize covariance chain, local mean, local sum

    Args:
        * **x** (:class:`~numpy.ndarray`): Chain segment
        * **w** (:class:`~numpy.ndarray`): Weights

    Returns:
        * **xcov** (:class:`~numpy.ndarray`): Initial covariance matrix
        * **xmean** (:class:`~numpy.ndarray`): Initial mean chain values
        * **wsum** (:class:`~numpy.ndarray`): Initial weighted sum
    '''
    npar = x.shape[1]
    wsum = sum(w)
    xmean = np.zeros(npar)
    xcov = np.zeros([npar,npar])
    for ii in range(npar):
        xmean[ii] = sum(x[:,ii]*w)*(wsum**(-1))
    if wsum > 1:
        for ii in range(0,npar):
            for jj in range(0,ii+1):
                term1 = x[:,ii] - xmean[ii]
                term2 = x[:,jj] - xmean[jj]
                xcov[ii,jj] = np.dot(term1.transpose(), ((term2)*w)*((wsum-1)**(-1)))
                if ii != jj:
                    xcov[jj,ii] = xcov[ii,jj]
    return xcov, xmean, wsum
# --------------------------------------------
def setup_cholupdate(R, oldwsum, wsum, oldmean, xi):
    '''
    Setup input arguments to the Cholesky update function

    Args:
        * **R** (:class:`~numpy.ndarray`): Previous Cholesky decomposition matrix
        * **oldwsum** (:class:`~numpy.ndarray`): Previous weighted sum
        * **w** (:class:`~numpy.ndarray`): Current Weights
        * **oldmean** (:class:`~numpy.ndarray`): Previous mean chain values
        * **xi** (:class:`~numpy.ndarray`): Row of chain segment

    Returns:
        * **Rin** (:class:`~numpy.ndarray`): Scaled Cholesky decomposition matrix
        * **xin** (:class:`~numpy.ndarray`): Updated mean chain values for Cholesky function
    '''
    Rin = np.sqrt((oldwsum-1)*((wsum+oldwsum-1)**(-1)))*R
    xin = (xi - oldmean).transpose()*np.sqrt(((wsum*oldwsum)*((wsum+oldwsum-1)**(-1))*((wsum+oldwsum)**(-1))))
    return Rin, xin
# --------------------------------------------
# Cholesky Update
def cholupdate(R, x):
    '''
    Update Cholesky decomposition

    Args:
        * **R** (:class:`~numpy.ndarray`): Weighted Cholesky decomposition
        * **x** (:class:`~numpy.ndarray`): Weighted sum based on local chain update

    Returns:
        * **R1** (:class:`~numpy.ndarray`): Updated Cholesky decomposition
    '''
    n = len(x)
    R1 = R.copy()
    x1 = x.copy()
    for ii in range(n):
        r = math.sqrt(R1[ii,ii]**2 + x1[ii]**2)
        c = r*(R1[ii,ii]**(-1))
        s = x1[ii]*(R1[ii,ii]**(-1))
        R1[ii,ii] = r
        if ii+1 < n:
            R1[ii,ii+1:n] = (R1[ii,ii+1:n] + s*x1[ii+1:n])*(c**(-1))
            x1[ii+1:n] = c*x1[ii+1:n] - s*R1[ii,ii+1:n]

    return R1
# --------------------------------------------
def update_cov_via_ram(u, isimu, etaparam, npar, alphatarget, alpha, R):
    '''
    Update covariance matrix via RAM

    Args:
        * **u** (:class:`~numpy.ndarray`): Latest random sample points
        * **isimu** (:py:class:`int`): Simulation counter
        * **alphatarget** (:py:class:`float`): Acceptance ratio target.
        * **npar** (:py:class:`int`): Number of parameters.
        * **etaparam** (:py:class:`float`):
        * **alpha** (:py:class:`float`): Latest Likelihood evaluation
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.

    Returns:
        * **upcov** (:class:`~numpy.ndarray`): Updated parameter covariance matrix.
    '''
    uu = u*(np.linalg.norm(u)**(-1))
    eta = (isimu**(etaparam))**(-1)
    ram = np.eye(npar) + eta*(min(1.0, alpha) - alphatarget)*(np.dot(uu.transpose(), uu))
    upcov = np.dot(np.dot(R.transpose(),ram),R)
    return upcov
# --------------------------------------------
def update_cov_from_covchain(covchain, qcov, no_adapt_index):
    '''
    Update covariance matrix from covariance matrix chain

    Args:
        * **covchain** (:class:`~numpy.ndarray`): Covariance matrix history.
        * **qcov** (:class:`~numpy.ndarray`): Parameter covariance matrix
        * **no_adapt_index** (:class:`numpy.ndarray`): Indices of parameters not being adapted.

    Returns:
        * **upcov** (:class:`~numpy.ndarray`): Updated covariance matrix
    '''
    upcov = covchain.copy()
    upcov[no_adapt_index, :] = qcov[no_adapt_index,:]
    upcov[:,no_adapt_index] = qcov[:,no_adapt_index]
    return upcov
# --------------------------------------------
def check_for_singular_cov_matrix(upcov, R, npar, qcov_adjust, qcov_scale, rejected, iiadapt, verbosity):
    '''
    Check if singular covariance matrix

    Args:
        * **upcov** (:class:`~numpy.ndarray`): Parameter covariance matrix.
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
        * **npar** (:py:class:`int`): Number of parameters.
        * **qcov_adjust** (:py:class:`float`): Covariance adjustment factor.
        * **qcov_scale** (:py:class:`float`): Scale parameter
        * **rejected** (:py:class:`dict`): Rejection counters.
        * **iiadapt** (:py:class:`int`): Adaptation counter.
        * **verbosity** (:py:class:`int`): Verbosity of display output.

    Returns:
        * **R** (:class:`~numpy.ndarray`): Adjusted Cholesky decomposition of covariance matrix.
    '''
    pos_def, pRa = is_semi_pos_def_chol(upcov)
    if pos_def == 1: # not singular!
        return scale_cholesky_decomposition(Ra = pRa, qcov_scale = qcov_scale)
    else: # singular covariance matrix
        return adjust_cov_matrix(upcov = upcov, R = R, npar = npar, qcov_adjust = qcov_adjust, qcov_scale = qcov_scale, rejected = rejected, iiadapt = iiadapt, verbosity = verbosity)
# --------------------------------------------
def is_semi_pos_def_chol(x):
    '''
    Check if matrix is semi-positive definite using Cholesky Decomposition

    Args:
        * **x** (:class:`~numpy.ndarray`): Covariance matrix

    Returns:
        * `Boolean`
        * **c** (:class:`~numpy.ndarray`): Cholesky decomposition (upper triangular form) or `None`
    '''
    
    c = None
    try:
        c = np.linalg.cholesky(x)
        return True, c.transpose()
    except np.linalg.linalg.LinAlgError:
        return False, c
# --------------------------------------------
def scale_cholesky_decomposition(Ra, qcov_scale):
    '''
    Scale Cholesky decomposition

    Args:
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
        * **qcov_scale** (:py:class:`float`): Scale factor

    Returns:
        * **R** (:class:`~numpy.ndarray`): Scaled Cholesky decomposition of covariance matrix.
    '''
    return Ra*qcov_scale
# --------------------------------------------
def adjust_cov_matrix(upcov, R, npar, qcov_adjust, qcov_scale, rejected, iiadapt, verbosity):
    '''
    Adjust covariance matrix if found to be singular.

    Args:
        * **upcov** (:class:`~numpy.ndarray`): Parameter covariance matrix.
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
        * **npar** (:py:class:`int`): Number of parameters.
        * **qcov_adjust** (:py:class:`float`): Covariance adjustment factor.
        * **qcov_scale** (:py:class:`float`): Scale parameter
        * **rejected** (:py:class:`dict`): Rejection counters.
        * **iiadapt** (:py:class:`int`): Adaptation counter.
        * **verbosity** (:py:class:`int`): Verbosity of display output.

    Returns:
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
    '''
    # try to blow it up
    tmp = upcov + np.eye(npar)*qcov_adjust
    pos_def_adjust, pRa = is_semi_pos_def_chol(tmp)
    if pos_def_adjust == 1: # not singular!
        message(verbosity, 1, 'adjusted covariance matrix')
        # scale R
        R = scale_cholesky_decomposition(Ra = pRa, qcov_scale = qcov_scale)
        return R
    else: # still singular...
        errstr = str('covariance matrix singular, no adaptation')
        message(verbosity, 0, '{} {}'.format(errstr, rejected['in_adaptation_interval']*(iiadapt**(-1))*100))
        return R
# --------------------------------------------
def update_delayed_rejection(R, npar, ntry, drscale):
    '''
    Update Cholesky/Inverse matrix for Delayed Rejection

    Args:
        * **R** (:class:`~numpy.ndarray`): Cholesky decomposition of covariance matrix.
        * **npar** (:py:class:`int`): Number of parameters.
        * **ntry** (:py:class:`int`): Number of tries to take before rejection. Default is method dependent.
        * **drscale** (:class:`~numpy.ndarray`): Reduced scale for sampling in DR algorithm. Default is [5,4,3].

    Returns:
        * **RDR** (:py:class:`list`): List of Cholesky decomposition of covariance matrices for each stage of DR.
        * **InvR** (:py:class:`list`): List of Inverse Cholesky decomposition of covariance matrices for each stage of DR.
    '''
    RDR = None
    invR = None
    if ntry > 1: # delayed rejection
        RDR = []
        invR = []
        RDR.append(R)
        invR.append(np.linalg.solve(RDR[0], np.eye(npar)))
        for ii in range(1,ntry):
            RDR.append(RDR[ii-1]*((drscale[min(ii,len(drscale)) - 1])**(-1)))
            invR.append(invR[ii-1]*(drscale[min(ii,len(drscale)) - 1]))
        
    return RDR, invR


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 17 09:06:51 2018

@author: prmiles
"""

# import required packages
import numpy as np
import sys
#import warnings

class ModelSettings:
    '''
    MCMC Model Settings

    Attributes:
        * :meth:`~define_model_settings`
        * :meth:`~display_model_settings`
    '''
    def __init__(self):
        # Initialize all variables to default values
        self.description = 'Model Settings'

    def define_model_settings(self, sos_function = None, prior_function = None, prior_type = 1,
                 prior_update_function = None, prior_pars = None, model_function = None,
                 sigma2 = None, N = None, S20 = np.nan, N0 = None, nbatch = None):
        '''
        Define model settings.

        Args:
            * **sos_function**: Handle for sum-of-squares function
            * **prior_function**: Handle for prior function
            * **prior_type**: Pending...
            * **prior_update_function**: Pending...
            * **prior_pars**: Pending...
            * **model_function**: Handle for model function (needed if :code:`sos_function` not specified)
            * **sigma2** (:py:class:`float`): List of initial error observations.
            * **N** (:py:class:`int`): Number of observations - see :class:`~.DataStructure`.
            * **S20** (:py:class:`float`): List of scaling parameter in observation error estimate.
            * **N0** (:py:class:`float`): List of scaling parameter in observation error estimate.
            * **nbatch** (:py:class:`int`): Number of batch data sets - see :meth:`~.get_number_of_batches`.

        .. note:: Variables :code:`sigma2, N, S20, N0`, and :code:`nbatch` converted to :class:`~numpy.ndarray` for subsequent processing.
        '''
    
        self.sos_function = sos_function
        self.prior_function = prior_function
        self.prior_type = prior_type
        self.prior_update_function = prior_update_function
        self.prior_pars = prior_pars
        self.model_function = model_function
        
        # check value of sigma2 - initial error variance
        self.sigma2 = self._array_type(sigma2)
        
        # check value of N - total number of observations
        self.N = self._array_type(N)
        
        # check value of N0 - prior accuracy for S20
        self.N0 = self._array_type(N0)
        
        # check nbatch - number of data sets
        self.nbatch = self._array_type(nbatch)
            
        # S20 - prior for sigma2
        self.S20 = self._array_type(S20)
    
    @classmethod
    def _array_type(cls, x):
        # All settings in this class should be converted to numpy ndarray
        if x is None:
            return None
        else:
            if isinstance(x, int): # scalar -> ndarray[scalar]
                return np.array([np.array(x)])
            elif isinstance(x, float): # scalar -> ndarray[scalar]
                return np.array([np.array(x)])
            elif isinstance(x, list): # [...] -> ndarray[...]
                return np.array(x)
            elif isinstance(x, np.ndarray):
                return x
            else:
                sys.exit('Unknown data type - Please use int, ndarray, or list')
    
    def _check_dependent_model_settings(self, data, options):
        '''
        Check dependent parameters.

        Args:
            * **data** (:class:`~.DataStructure`): MCMC data structure
            * **options** (:class:`.SimulationOptions`): MCMC simulation options
        '''
        if self.nbatch is None:
            self.nbatch = data.get_number_of_batches()
            
        if self.N is not None:
            self.N = self._check_number_of_observations(udN = self.N, dsN = self._array_type(data.n))
        else:
#            self.N = data.get_number_of_observations()
            self.N = data.n
        
        self.Nshape = data.shape
        
        # This is for backward compatibility
        # if sigma2 given then default N0=1, else default N0=0
        if self.N0 is None:
            if self.sigma2 is None:
                self.sigma2 = np.ones([1])
                self.N0 = np.zeros([1])
            else:
                self.N0 = np.ones([1])
            
        # set default value for sigma2
        # default for sigma2 is S20 or 1
        if self.sigma2 is None:
            if not(np.isnan(self.S20)).any:
                self.sigma2 = self.S20
            else:
                self.sigma2 = np.ones(self.nbatch)
        
        if np.isnan(self.S20).any():
            self.S20 = self.sigma2  # prior parameters for the error variance
      
    @classmethod
    def _check_number_of_observations(cls, udN, dsN):
        # check if user defined N matches data structure
        if np.array_equal(udN, dsN):
            N = dsN
        elif dsN.size > udN.size and udN.size == 1:
            if np.all(dsN == udN[0]):
                N = dsN
            else:
                sys.exit('User defined N = {}.  Estimate based on data structure is N = {}.  Possible error?'.format(udN, dsN))
        elif udN.size > dsN.size and dsN.size == 1:
            if np.all(dsN[0] == udN):
                N = udN
            else:
                sys.exit('User defined N = {}.  Estimate based on data structure is N = {}.  Possible error?'.format(udN, dsN))
        else:
            sys.exit('User defined N = {}.  Estimate based on data structure is N = {}.  Possible error?'.format(udN, dsN))
                
        return N
    
    def _check_dependent_model_settings_wrt_nsos(self, nsos):
        '''
        Check dependent model settings with respect to number of sum-of-square elements.

        A different observation error can be set up for different data sets, so these
        settings must be updated with respect to the length of the output from the
        sum-of-squares function evaluation.

        Args:
            * **nsos** (:py:class:`int`): Length of output from sum-of-squares function
        '''
        # in matlab version, ny = length(sos) where sos is the output from the sos evaluation
        self.S20 = self.__check_size_of_setting_wrt_nsos(self.S20, nsos)
        self.N0 = self.__check_size_of_setting_wrt_nsos(self.N0, nsos)
        self.sigma2 = self.__check_size_of_setting_wrt_nsos(self.sigma2, nsos)
        self.N = self.__check_size_of_observations_wrt_nsos(self.N, nsos, self.Nshape)
#        if len(self.N) == 1:
#            self.N = np.ones(nsos)*self.N
            
#        if len(self.N) == nsos + 1:
#            self.N = self.N[1:] # remove first column
            
        self.nsos = nsos
        
    @classmethod
    def __check_size_of_setting_wrt_nsos(cls, x, nsos):
        '''
        Check size of setting with respect number of observation errors.

        Args:
            * **x** (:class:`~numpy.ndarray`): Array to be checked
            * **nsos** (:py:class:`int`): Length of output from sum-of-squares function

        Returns:
            * **x** (:class:`~numpy.ndarray`): Array returned with shape = :code:`(nsos,1)`

        Raises:
            * Dimension mismatch if :code:`len(x) > nsos` or :code:`len(x) < nsos and len(x) != 1`.
        '''
        if len(x) == 1:
            x = np.ones(nsos)*x
        elif len(x) > nsos: # more x elements than sos elements
            sys.exit('SOS function returns nsos = {} elements.  Expect same number of elements in S20, N0, sigma2'.format(nsos))
        elif len(x) < nsos and len(x) != 1:
            sys.exit('SOS function returns nsos = {} elements.  Length of S20, N0, or sigma2 is {}.  These settings must have one element or nsos elements.'.format(nsos, len(x)))
        return x
     
    @classmethod
    def __check_size_of_observations_wrt_nsos(cls, N, nsos, Nshape):
        if len(N) == 1:
            N = np.ones(nsos)*N
        elif len(N) < nsos and len(N) > 1:
            sys.exit('Unclear data structure!  SOS function return nsos = {} elements.  len(N) = {}.'.format(nsos, len(N)))
        elif nsos == 1 and len(N) > 1:
            N = np.array([np.sum(N)])
            
        return N
            
    def display_model_settings(self, print_these = None):
        '''
        Display subset of the simulation options.

        Args:
            * **print_these** (:py:class:`list`): List of strings corresponding to keywords.  Default below.

        ::

            print_these = ['sos_function', 'model_function', 'sigma2', 'N', 'N0', 'S20', 'nsos', 'nbatch']
        '''
        if print_these is None:
            print_these = ['sos_function', 'model_function', 'sigma2', 'N', 'N0', 'S20', 'nsos', 'nbatch']
            
        print('model settings:')
        for ptii in print_these:
            print('\t{} = {}'.format(ptii, getattr(self, ptii)))
            
        return print_these